#include <KD/kd.h>
#include "vbo.h"
#include "buildtree.h"
#include "branches.h"
#include "shaders.h"

static Array vertices;
static Array normals;
static Array texcoords;

static unsigned int VBOvertices;
static unsigned int VBOnormals;
static unsigned int VBOtexcoords;

static Array indices;
static GLuint texture;

static float2 geom[BRANCHES_FACETS + 1];

void Branches_initialize(GLuint t)
{
    int i;

    Array_init(&vertices, sizeof(float3));
    Array_init(&normals, sizeof(float3));
    Array_init(&texcoords, sizeof(float2));
    Array_init(&indices, sizeof(unsigned int));

    texture = t;

    if (use_VBO) {
        VBOvertices = 0;
        VBOnormals = 0;
        VBOtexcoords = 0;
    }

    for (i=0; i< BRANCHES_FACETS + 1; ++i) {
        float u = (float)(2.0f * KD_PI_F / (float)BRANCHES_FACETS) * (float)i;
        set_2(geom[i], kdCosf(u), kdSinf(u));
    }
}


void Branches_deinitialize()
{
    Array_destroy(&vertices);
    Array_destroy(&normals);
    Array_destroy(&texcoords);
    Array_destroy(&indices);
}


void Branches_clear()
{
    Array_clear(&vertices);
    Array_clear(&normals);
    Array_clear(&texcoords);
    Array_clear(&indices);
}


int Branches_add(float3 n, float2 tc, float3 v) {

    int index = vertices.elem_count;
    kdAssert(normals.elem_count == index);
    kdAssert(texcoords.elem_count == index);

    Array_push(&normals, n);
    Array_push(&texcoords, tc);
    Array_push(&vertices, v);

    return index;
}


void Branches_addIndex(unsigned int i)
{
   Array_push(&indices, &i);
}


void Branches_draw(int is_VBO)
{
    int i, size, stride;

    glUseProgram(prog_solids);

    glVertexAttrib3f(aloc_solidsColor, 1.0,1.0,1.0);
    glEnableVertexAttribArray(aloc_solidsVertex);
    glEnableVertexAttribArray(aloc_solidsNormal);
    glEnableVertexAttribArray(aloc_solidsTexcoord);

    if (is_VBO) {
        glBindBuffer(GL_ARRAY_BUFFER, VBO_NAME);
        glVertexAttribPointer(aloc_solidsVertex,
                              3, GL_FLOAT, GL_FALSE, 0, (void *)VBOvertices);
        glVertexAttribPointer(aloc_solidsNormal,
                              3, GL_FLOAT, GL_FALSE, 0, (void *)VBOnormals);
        glVertexAttribPointer(aloc_solidsTexcoord,
                              2, GL_FLOAT, GL_FALSE, 0, (void *)VBOtexcoords);
    } else {
        glVertexAttribPointer(aloc_solidsVertex,
                              3, GL_FLOAT, GL_FALSE, 0, vertices.buffer);
        glVertexAttribPointer(aloc_solidsNormal,
                              3, GL_FLOAT, GL_FALSE, 0, normals.buffer);
        glVertexAttribPointer(aloc_solidsTexcoord,
                              2, GL_FLOAT, GL_FALSE, 0, texcoords.buffer);

    }

    size = indices.elem_count;
    stride = (BRANCHES_FACETS + 1) * 2;
    kdAssert(size % stride == 0);

    glBindTexture(GL_TEXTURE_2D, texture);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    for (i=0; i<size; i+=stride) {
        glDrawElements(GL_TRIANGLE_STRIP, stride, GL_UNSIGNED_INT,
                       Array_get(&indices, i));
    }

    if (is_VBO) {
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }

    glDisableVertexAttribArray(aloc_solidsVertex);
    glDisableVertexAttribArray(aloc_solidsNormal);
    glDisableVertexAttribArray(aloc_solidsTexcoord);
}


int Branches_polyCount() {
    int stride = (BRANCHES_FACETS+1)*2;
    int cyl_count = indices.elem_count/stride;
    return cyl_count * BRANCHES_FACETS * 2;
}


int Branches_branchCount() {
    int stride = (BRANCHES_FACETS+1)*2;
    int cyl_count = indices.elem_count/stride;
    return (cyl_count+1) / 2;
}

void Branches_buildVBO()
{
    int v;

    v = vertices.elem_count;

    glBindBuffer(GL_ARRAY_BUFFER, VBO_NAME);
    VBOvertices  = VBO_alloc(v * 3 * sizeof(float));
    VBOnormals   = VBO_alloc(v * 3 * sizeof(float));
    VBOtexcoords = VBO_alloc(v * 2 * sizeof(float));

    glBufferSubData(GL_ARRAY_BUFFER, VBOvertices, v * 3 * sizeof(float),
                    vertices.buffer);
    glBufferSubData(GL_ARRAY_BUFFER, VBOnormals, v * 3 * sizeof(float),
                    normals.buffer);
    glBufferSubData(GL_ARRAY_BUFFER, VBOtexcoords, v * 2 * sizeof(float),
                    texcoords.buffer);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

int Branches_sizeVBO()
{
    int numverts = Branches_numVertices();
    return
        VBO_align(numverts*3*sizeof(GLfloat)) +   // normals
        VBO_align(numverts*2*sizeof(GLfloat)) +   // texcoords
        VBO_align(numverts*3*sizeof(GLfloat));    // verts.
}

int Branches_numVertices() { return vertices.elem_count; }


void Branches_generateStump(int *lower)
{
    int i;
    float branch_radius;

    for(i = 0; i < BRANCHES_FACETS+1; ++i)
    {
        float t = 1.0f / BRANCHES_FACETS * i;
        float g0 = geom[i][0];
        float g1 = geom[i][1];

        Branches_addIndex(lower[i]);
        branch_radius = tree_params[TREE_PARAM_BRANCH_SIZE];
        {
            float3 n = {g0, g1, 0.5f};
            float2 tc = {t, -branch_radius - 0.5f};
            float3 v = {g0 * branch_radius * 1.5f,
                        g1 * branch_radius * 1.5f,
                        -0.5f};
            Branches_addIndex(Branches_add(n, tc, v));
        }
    }
}


void Branches_buildCylinder(int *idx, float4x4 mat, float taper,
                            float texcoord_y, GLboolean low)
{
    int i;
    for (i=0; i<(BRANCHES_FACETS+1); ++i)
    {
        float t = 1.0f / BRANCHES_FACETS * i;
        float3 n, v;
        float *g = geom[i];
        float branch_radius;

        set_3(n, g[0], g[1], 0.0f);
        transformVeci_f3(n, mat);
        branch_radius = tree_params[TREE_PARAM_BRANCH_SIZE];
        if(low)
        {
            set_3(v, g[0] * branch_radius,
                     g[1] * branch_radius,
                     branch_radius);
        }
        else
        {
            set_3(v, g[0] * branch_radius * taper,
                     g[1] * branch_radius * taper,
                     1.0f - branch_radius);
        }
        transformi_f3(v, mat);

        {
            float2 tc = {t, texcoord_y};
            idx[i] = Branches_add(n, tc, v);
        }
    }
}
