#include <KD/kd.h>

#include "vector.h"
#include "random.h"

#include "buildtree.h"
#include "tree.h"
#include "branches.h"
#include "leaves.h"


////////////////////////////////////////////////////////////
// provide a function which generates the tree.


static float treebuild_threshhold;



// We keep the random numbers at each branch.  This mini-object
// keeps track of those random numbers.

typedef struct BranchNoise {
  float noise;

  struct BranchNoise* left;
  struct BranchNoise* right;

} BranchNoise;

// initialization and clean-up.
static BranchNoise *BranchNoise_new(void)
{
   BranchNoise *o = (BranchNoise *)kdMalloc(sizeof(BranchNoise));
   o->left = KD_NULL;
   o->right = KD_NULL;

   o->noise = ((float) get_random()) * 0.3f - 0.1f;

   return o;
}
static void BranchNoise_delete(BranchNoise *o)
{
    kdAssert(o);

    // Recursively delete all the branches.
    if(o->left) { BranchNoise_delete(o->left); }
    if(o->right) { BranchNoise_delete(o->right); }

    kdFree(o);
}


// provide access to left and right branches of the noise tree.
// if the branch doesn't exist, it'll be created.
static BranchNoise* BranchNoise_left(BranchNoise *o)
{
    kdAssert(o);
    if (!o->left) { o->left = BranchNoise_new(); }
    kdAssert(o->left);
    return o->left;
}
static BranchNoise* BranchNoise_right(BranchNoise *o)
{
    kdAssert(o);
    if (!o->right) { o->right = BranchNoise_new(); }
    kdAssert(o->right);
    return o->right;
}


// We only need one root BranchNoise object at a time and this is the one.
static BranchNoise *bn = KD_NULL;



////////////////////////////////////////////////////////////
// A recursive routine to generate a tree.

// since we have indirect recursive calls, we need a forward declaration.
static void build(int lower[BRANCHES_FACETS + 1],
        BranchNoise *noise, float4x4 mat,
        float texcoord_y, float decay, int level);

// this one is called once per branch.
static void buildBranch(float radius, float angle, float twist, float decay, int level,
        float texcoord_y, float4x4 translate_mat, BranchNoise *branch_bn,
        int upper[])
{
    float4x4 mat, scale_mat, rot_mat;
    float dec;
    int i;

    makeScale(scale_mat, radius, radius, radius),
    makeRotation(rot_mat, kdSinf(twist), kdCosf(twist), 0.0f, angle);
    mult_f4x4(mat, scale_mat, rot_mat);
    multi_f4x4(mat, translate_mat);

    dec = radius * decay;

    if((level + 1) >= BRANCH_DEPTH || dec < treebuild_threshhold)
    {
        Leaves_add(mat);
    }
    else
    {
        int lower[BRANCHES_FACETS + 1];

        build(lower, branch_bn, mat, texcoord_y, dec, level + 1);

        for (i = 0; i < BRANCHES_FACETS + 1; ++i)
        {
            Branches_addIndex(lower[i]);
            Branches_addIndex(upper[i]);
        }
    }
}


static void build(int lower[BRANCHES_FACETS + 1],
        BranchNoise *noise, float4x4 mat, float texcoord_y, float decay, int level)
{
    int upper[BRANCHES_FACETS + 1];
    float btwist, left_branch_noise, right_branch_noise;
    float branch_angle, branch_angle_bias;
    float left_radius, left_angle, right_radius, right_angle;
    float taper, branch_radius;
    BranchNoise *left_bn, *right_bn;
    float4x4 translate_mat;
    int i;

    kdAssert(noise);

    Leaves_setRadius(tree_params[TREE_PARAM_LEAF_SIZE]);

    btwist = tree_params[TREE_PARAM_TWIST] * (level + 1);

    left_bn = BranchNoise_left(noise);
    right_bn = BranchNoise_right(noise);

    left_branch_noise = left_bn->noise * tree_params[TREE_PARAM_FULLNESS];
    right_branch_noise = right_bn->noise * tree_params[TREE_PARAM_FULLNESS];

    branch_angle = tree_params[TREE_PARAM_BALANCE];
    branch_angle_bias = tree_params[TREE_PARAM_SPREAD];

    left_radius = kdSqrtf(1.0 - branch_angle) + left_branch_noise;
    left_radius = clamp(left_radius, 0.0f, 1.0f);
    left_angle = (float)(branch_angle * branch_angle_bias * KD_PI_F / 2.0f);

    right_radius = kdSqrtf(branch_angle) + right_branch_noise;
    right_radius = clamp(right_radius, 0.0f, 1.0f);
    right_angle = (float)((branch_angle - 1.0f) * branch_angle_bias * KD_PI_F / 2.0f);

    taper = (left_radius > right_radius) ? left_radius : right_radius;

    branch_radius = tree_params[TREE_PARAM_BRANCH_SIZE];

    Branches_buildCylinder(lower, mat, taper, texcoord_y, GL_TRUE);
    texcoord_y += 1.0f - 2 * branch_radius;
    Branches_buildCylinder(upper, mat, taper, texcoord_y, GL_FALSE);
    texcoord_y += 2 * branch_radius;

    makeTranslate(translate_mat, 0.0f, 0.0f, 1.0f);
    multi_f4x4(translate_mat, mat);

    for (i = 0; i < BRANCHES_FACETS + 1; ++i)
    {
        Branches_addIndex(upper[i]);
        Branches_addIndex(lower[i]);
    }
    buildBranch(left_radius, left_angle, btwist,
            decay, level, texcoord_y, translate_mat, left_bn, upper);
    buildBranch(right_radius, right_angle, btwist,
            decay, level, texcoord_y, translate_mat, right_bn, upper);
}


void BuildTree_generate()
{
    int lower[BRANCHES_FACETS + 1];
    float u, max, min;

    // compute the threshhold.
    u = 1.0f - tree_params[TREE_PARAM_DEPTH];
    u = u * u * u * u;
    max = 0.5f;
    min = 0.03f;
    treebuild_threshhold = (max - min) * u + min;

    // Create the noise object if it's not there yet.
    if(!bn) { bn = BranchNoise_new(); }

    // Build the tree branches.
    build(lower, bn, ident_matrix_f, 0.0f, 1.0f, 0);

    // Build the tree stump.
    Branches_generateStump(lower);
}



void BuildTree_newCharacter()
{
    if(bn) { BranchNoise_delete(bn); bn = KD_NULL; }
}
