/*
 * main.c
 *
 * Copyright 2003-2009, NVIDIA Corporation.
 * All rights Reserved.
 *
 * THE INFORMATION CONTAINED HEREIN IS PROPRIETARY AND CONFIDENTIAL TO
 * NVIDIA, CORPORATION.  USE, REPRODUCTION OR DISCLOSURE TO ANY THIRD PARTY
 * IS SUBJECT TO WRITTEN PRE-APPROVAL BY NVIDIA, CORPORATION.
 */

// Implements a simple EGL main loop.

#include <KD/kd.h>
#include "nvgl2demo_common.h"
#include "screen.h"

static GLboolean shutdown = GL_FALSE;

// locate a substring in a null terminated string, return NULL if not found
static char* CtreeStrStr(const char* str1, const char* str2)
{
    int len1;
    int len2;
    KDchar* start;

    if (!str1 || !str2)
    {
        return KD_NULL;
    }

    len1 = kdStrlen(str1);
    len2 = kdStrlen(str2);

    // Find the first occurence
    start = kdStrchr(str1, *str2);

    // If we found the first character, continue to test
    // See if is possible to fit str2 into the rest of str1
    while ( start ) {

        if (!kdStrncmp(str2, start, len2)) {
            return start;
        }

        // Find the first occurence
        start = kdStrchr(start + 1, *str2);
    }

    // If we got here, we didn't find the string
    return KD_NULL;
}

// Callback to close window
static void
closeCB(void)
{
    shutdown = GL_TRUE;
}

// Callback to resize window
static void
resizeCB(int width, int height)
{
    Screen_resize(width, height);
}

// Callback to handle key presses
static void
keyCB(char key, int state)
{
    // Ignoring releases
    if (!state) return;

    Screen_callback(key, 0, 0);
}

// Entry point of this demo program.
KDint
kdMain(KDint argc, const KDchar *const *argv)
{
    GLboolean fpsFlag = GL_FALSE;
    int setDemoParams = GL_FALSE;
    float duration = 0.0f;
    int width, height;


    // Initialize window system and EGL
    if (!NvGl2DemoInitialize(&argc, (char**) argv, "ctree",
                             EGL_OPENGL_ES2_BIT)) {
        return -1;
    }

    // Set up callbacks
    NvGl2DemoSetCloseCB(closeCB);
    NvGl2DemoSetResizeCB(resizeCB);
    NvGl2DemoSetKeyCB(keyCB);

    // Parse the command line options.
    argc--;
    argv++;
    while (argc > 0) {
        if (!kdStrcmp(argv[0], "-sec")) {
            if (argc >= 2) {
                duration = (float) kdStrtof(argv[1], KD_NULL);
                argc -= 2;
                argv += 2;
            } else {
                NvGl2DemoLogMessage("Insufficient parameters to <%s>\n", argv[0]);
                goto done;
            }
        } else if (!kdStrcmp(argv[0], "-demo")) {
            setDemoParams = 1;
            argc -= 1;
            argv += 1;
        } else if (!kdStrcmp(argv[0], "-smalltex")) {
            Screen_setSmallTex();
            argc -= 1;
            argv += 1;
        } else if (!kdStrcmp(argv[0], "-nomenu")) {
            Screen_setNoMenu();
            argc -= 1;
            argv += 1;
        } else if (!kdStrcmp(argv[0], "-nosky")) {
            Screen_setNoSky();
            argc -= 1;
            argv += 1;
        } else if (!kdStrcmp(argv[0], "-fps")) {
            fpsFlag = GL_TRUE;
            argc -= 1;
            argv += 1;
        } else {
            NvGl2DemoLogMessage("Unknown command line option <%s>\n", argv[0]);
            goto done;
        }
    }

    (void)CtreeStrStr; // Fix compiler warning
#if 0
    // Check for required extensions.
    if (!CtreeStrStr((const char *)glGetString(GL_EXTENSIONS),
                "OES_element_index_uint ")) {
        NvGl2DemoLogMessage("Required extension OES_element_index_uint is missing!\n");
        goto done;
    }
#endif

    width = gfxState.width;
    height = gfxState.height;

    // Initialize the main window.
    Screen_initialize(duration, width, height, fpsFlag);
    if (setDemoParams) { Screen_setDemoParams(); }

    // Main loop.
    do {
        // Rendering.
        Screen_draw();
        eglSwapBuffers(gfxState.display, gfxState.surface);
        NvGl2DemoCheckEvents();
    } while (!Screen_isFinished() && !shutdown);

    // Clean up the main window.
    Screen_deinitialize();

    done:

    // Clean up EGL
    NvGl2DemoShutdown();

    return 0;
}
