#include <KD/kd.h>
#include <KD/KHR_formatted.h>

#include "nvgl2demo_common.h"

#include "array.h"
#include "random.h"
#include "vector.h"
#include "buildtree.h"
#include "screen.h"

#include "nvtexfont2.h"

#include "shaders.h"

#include "vbo.h"


//NOTE: any changes to NUM_LIGHTS must also be made to vert.glslv
#define NUM_LIGHTS 8

#ifdef DUMP_VERTS_PER_SEC
int vertCount = 0;
#endif

// Texture font.
NVTexfont2RasterFont *nvtxf = KD_NULL;

// Font height, as a ratio of the screen height
//#define FONT_HEIGHT_RATIO (1.0f/33.0f)

// We store picture's info into this structure.
static struct pictInfoStruct {
   char *filename;
   float left, right, bottom, top;
} pictInfo[] = {
   {"textures/label_depth.tga", 0, 150, 260, 275},
   {"textures/label_balance.tga", 0, 150, 220, 235},
   {"textures/label_twist.tga", 0, 150, 180, 195},
   {"textures/label_spread.tga", 0, 150, 140, 155},
   {"textures/label_leaf_size.tga", 0, 150, 100, 115},
   {"textures/label_branch_size.tga", 0, 150, 60, 75},
   {"textures/label_fullness.tga", 0, 150, 20, 35},
   // We're projecting all controls in ortho VGA (640x480) anyway,
   // so this should always be up-left corner of the picture.
   {"textures/NVidiaLogo.tga", 0, 150, 450, 480},
};
#define NUM_PICTS (sizeof(pictInfo) / sizeof(struct pictInfoStruct))
static int const LOGO_PICT = NUM_PICTS - 1;


// TreePos object.
typedef struct
{
  float x;
  float y;
  float angle;
  Firefly fireflies[NUM_LIGHTS];
} TreePos;

static TreePos *TreePos_new(float nx, float ny, float a)
{
    int i;
    TreePos *o = (TreePos *)kdMalloc(sizeof(TreePos));
    o->x = nx;
    o->y = ny;
    o->angle = a;

    for(i = 0; i < NUM_LIGHTS; i++)
    {
        Firefly_init(o->fireflies + i, i);
    }

    return o;
}
static void TreePos_delete(TreePos *o)
{
    int i;
    for(i = 0; i < NUM_LIGHTS; i++)
    {
        Firefly_destroy(o->fireflies + i);
    }
    kdFree(o);
}

//////////////////////////////////////////////////////////////////////
// variables used in this module.

// Properties of the window.
static int width, height;
static int swapInterval = 0;

// light_count is the current number of lights (bugs) where
// NUM_LIGHTS is the max number of lights.
static int light_count = 3;

// camera properties.
static float3 eye = {0.0f, 1.0f, 7.0f};
static float heading = 0.0f;
static float pitch = 17.0f;
static float dh = 0.0f;
static float dp = 0.0f;
static float velocity = 0.0f;

// keep track of all the trees in the scene.
static Array tree_poses;

// 2D items in the window.
static Slider *sliders[NUM_TREE_PARAMS];
static int selectedSlider = TREE_PARAM_DEPTH;
static Picture *picts[8];
static GLboolean overlay_flag = GL_TRUE;

//If this flag is true, FPS is visible on screen
static GLboolean fps_flag = GL_TRUE;

//If this flag is true,
//FPS is displayed/logged on the text console/terminal (only once)
static GLboolean fps_log_flag = GL_TRUE;

// keeping track of the time.
#define FPS_PERIOD 500 // How many frames to recompute fps
static double startTime;
static double currentTime;
static double duration;
static double fps;
static double dt = 1.0 / 20.0;

static int smalltex = 0;
static int nosky = 0;
static int nomenu = 0;

//////////////////////////////////////////////////////////////////////
// utility functions.

static void setSlider(int selectedSlider, float val, GLboolean mustSet)
{
    if(Slider_setValue(sliders[selectedSlider], val) || mustSet)
    {
        Tree_setParam(selectedSlider, val);
    }
}


static void tick(void)
{
    static int    frameCount = 0;
    static double lastTime = 0.0f;
    static double lastFpsTime;
    double        t;
    double        new_dt;

    // find dt, time took for one frame.
    t      = (double)kdGetTimeUST() / ((KDint64)1000*1000000);
    if (!lastTime) lastTime = lastFpsTime = t;
    new_dt = t - lastTime;
    if (new_dt<1.0f) {
        // smooth it out a little bit.
        dt = dt * 0.7f + new_dt * 0.3f;
    }
    //This line prevents from displaying fps = 0.0 till the initial FPS PERIOD
    if(!fps && frameCount >= 10) fps = 1.0 / new_dt;
    lastTime = t;

    // in theory, 1 / dt is the fps, but that changes too often.
    // we find out fps only once in time_frames frames.
    if (++frameCount == FPS_PERIOD)
    {
        frameCount = 0;
        fps = FPS_PERIOD / (t - lastFpsTime);

        if(fps_log_flag)
        {
            NvGl2DemoLogMessage("fps: %f\n", fps);
            fps_log_flag = GL_FALSE;
        }

#       ifdef DUMP_VERTS_PER_SEC
        {
            static float vertRate = 0;
            vertRate = (double)vertCount / (t - lastFpsTime);
            vertCount = 0;
            kdLtostr(buffer, 20, vertRate);
            kdStrncat_s(buffer, 20, " Verts/sec", 10);
            NvGl2DemoLogMessage(buffer);
        }
#       endif
        lastFpsTime = t;
    }
    currentTime = t;
}


static void refreshLights(void)
{
    glUseProgram(prog_solids);
    glUniform1i(uloc_solidsLights, light_count);
    glUseProgram(prog_leaves);
    glUniform1i(uloc_leavesLights, light_count);
}

//////////////////////////////////////////////////////////////////////
// initialize and de-initialize the module.

void Screen_initialize(float d, GLsizei w, GLsizei h, GLboolean fpsFlag)
{
    int i, ypos;
    TreePos *treeposPtr;

    fps_flag = fpsFlag;

    // Load shaders
    if (!LoadShaders()) return;

    // Initialize the module parameters.
    width  = w;
    height = h;

    //
    // 3D components initialization.
    //

    // Initialize fireflies
    Firefly_global_init(NUM_LIGHTS);

    // Initialize trees
    Tree_initialize(
    NvGl2DemoLoadTga(smalltex ? "textures/bark_s.tga" : "textures/bark.tga"),
    NvGl2DemoLoadTga(smalltex ? "textures/leaf_s.tga" : "textures/leaf.tga"),
    NvGl2DemoLoadTga(smalltex ? "textures/leaf_back_s.tga" : "textures/leaf_back.tga"));
    Array_init(&tree_poses, sizeof(TreePos));
    treeposPtr = TreePos_new(0.0f, 0.0f, 0.0f);
    Array_push(&tree_poses, treeposPtr);
    TreePos_delete (treeposPtr);

    // Initialize sky
    if (!nosky) {
        Sky_initialize(NvGl2DemoLoadTga("textures/sky_night.tga"));
    }

    // Initialize ground
    Ground_initialize(NvGl2DemoLoadTga(smalltex ? "textures/ground_s.tga" : "textures/ground.tga"));

    // Time tracking initialization.
    startTime = currentTime = (double)kdGetTimeUST() / ((KDint64)1000*1000000);
    duration  = d;
    tick();

    // Slider initialization.
    if (!nomenu) {
        for (i = 0, ypos = 240; i < NUM_TREE_PARAMS; i++, ypos -= 40)
        {
            sliders[i] = Slider_new(tree_params_min[i],
                                    tree_params_max[i],
                                    tree_params[i]);
            Slider_setPos(sliders[i],
                          0.0f, 150.0f,
                          (float) ypos, (float) (ypos + 20));
        }
        Slider_select(sliders[selectedSlider], GL_TRUE);
    }

    // Picture initialization.
    for(i = (nomenu ? LOGO_PICT : 0); i < NUM_PICTS; i++)
    {
        picts[i] = Picture_new(NvGl2DemoLoadTga(pictInfo[i].filename));
        Picture_setPos(picts[i],
                       pictInfo[i].left, pictInfo[i].right,
                       pictInfo[i].bottom, pictInfo[i].top);
    }

    // Some GL initializations here.
    eglSwapInterval(gfxState.display, swapInterval);
    glViewport(0, 0, width, height);
    glClearDepthf(1.0f);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glDepthRangef(0.0f, 1.0f);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    refreshLights();

    // All shaders use texture unit 0
    glUseProgram(prog_solids);
    glUniform1i(uloc_solidsTexUnit, 0);
    glUseProgram(prog_leaves);
    glUniform1i(uloc_leavesTexUnit, 0);
    glUseProgram(prog_simpletex);
    glUniform1i(uloc_simpletexTexUnit, 0);
    glUseProgram(prog_overlaytex);
    glUniform1i(uloc_overlaytexTexUnit, 0);

    nvtxf = nvtexfont2InitRasterFont(NV_TEXFONT_DEFAULT, 0, GL_TRUE,
                                     GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
}


void Screen_resize(GLsizei w, GLsizei h)
{
    width  = w;
    height = h;
    glViewport(0, 0, width, height);
}


void Screen_deinitialize()
{
    int i;
    for (i = (nomenu ? LOGO_PICT : 0); i < NUM_PICTS; i++) {
        Picture_delete(picts[i]);
    }
    if (!nomenu) {
        for (i = 0; i < NUM_TREE_PARAMS; i++) {
            Slider_delete(sliders[i]);
        }
    }
    Tree_deinitialize();

    Ground_deinitialize();

    if (!nosky) {
        Sky_deinitialize();
    }
    Array_destroy(&tree_poses);

    Firefly_global_destroy();

    if (nvtxf != KD_NULL)
        nvtexfont2UnloadRasterFont(nvtxf);
}


static void addTree(void)
{
    float scale = SKY_RADIUS - GROUND_SIZE;
    static float cone = 60.0f;

    float r = ((float) get_random()) * ((float) get_random()) * scale;
    float h = degToRadF(heading + (((float) get_random()) * cone - cone / 2.0f));
    TreePos *treeposPtr;
    treeposPtr = TreePos_new (kdCosf(h) * r, kdSinf(h) * r, ((float) get_random()) * 360.0f);
    Array_push(&tree_poses, treeposPtr);
    TreePos_delete (treeposPtr);
}

static GLboolean sliderControl(int key)
{
    if (nomenu) {
        return GL_TRUE;
    }
    switch (key)
    {
    case 'i':
    case 'k':
        {
        int dir = (key == 'i' ? -1 : 1);
        Slider_select(sliders[selectedSlider], GL_FALSE);
        selectedSlider += dir;
        if (selectedSlider == -1) { selectedSlider = NUM_TREE_PARAMS - 1; }
        if (selectedSlider == NUM_TREE_PARAMS) { selectedSlider = 0; }
        Slider_select(sliders[selectedSlider], GL_TRUE);
        return GL_TRUE;
        }

    case 'j':
    case 'l':
        {
        float val;
        val = tree_params_max[selectedSlider]-tree_params_min[selectedSlider];
        val *= 0.02f * (key == 'j' ? -1.0f : +1.0f);
        val += Slider_getValue(sliders[selectedSlider]);
        val = clamp(val, tree_params_min[selectedSlider],
                         tree_params_max[selectedSlider]),
        setSlider(selectedSlider, val, GL_FALSE);
        return GL_TRUE;
        }
    }
    return GL_FALSE;
}


static GLboolean cameraControl(int key)
{
    switch(key)
    {
    case 'j':
    case 'l':
        {
        float h = degToRadF(heading);
        float ch = kdCosf(h);
        float sh = kdSinf(h);
        float3 right_vec = {ch, 0.0f, sh};
        float3 tmp;
        mult_f3f(tmp, right_vec, 3.0f);
        multi_f3f(tmp, (float)(key == 'j' ? dt : -dt));
        addi_f3(eye, tmp);
        return GL_TRUE;
        }

    case 'i':
    case 'k':
        velocity += key == 'i' ? 0.2f : -0.2f;
        return GL_TRUE;

    case 'I':
    case 'K':
        eye[1] += (float)(key == 'I' ? dt : -dt);
        return GL_TRUE;
    }
    return GL_FALSE;
}

static GLboolean sceneControl(int key)
{
    switch (key)
    {
    case ' ':
        if (!nomenu) {
            overlay_flag = !overlay_flag;
        }
        return GL_TRUE;

    case 'q': case 'Q':
        // artifitially set the start time so we'll terminate.
        if(duration <= 0.0) { duration = 0.1; }
        startTime = currentTime - duration;
        return GL_TRUE;

    case 'h': case 'H':
        NvGl2DemoLogMessage("\n"
            "  SPACE: toggles the controls overlay\n"
            "\n"
            "  If the controls overlay is on:\n"
            "    i : select next higher parameter bar\n"
            "    j : decrease current parameter\n"
            "    k : select next lower parameter bar\n"
            "    l : increase current parameter\n"
            "\n"
            "  If the controls overlay is off\n"
            "    i : move forwards\n"
            "    j : move left\n"
            "    k : move backwards\n"
            "    l : move right\n"
            "    I : move up\n"
            "    K : move down\n"
            "\n"
            "  c    : change \"character\" of the tree\n"
            "  f    : toggle frames/sec display\n"
            "  S    : output polygon count to terminal window\n"
            "  v    : toggle wait for vertical sync on display device\n"
            "  1-8  : number of fireflies (colored point lights)\n"
            "  r    : toggle use of VBO\n"
            "  q    : quit\n"
            "\n");
        return GL_TRUE;

    case '+':
        addTree();
        return GL_TRUE;

    case '-':
        if (tree_poses.elem_count>1) { Array_pop(&tree_poses); }
        return GL_TRUE;

    case 'c':
        Tree_newCharacter();
        return GL_TRUE;

    case 'f':
        if (!nomenu) {
            fps_flag = !fps_flag;
        }
        fps_log_flag = GL_TRUE;
        return GL_TRUE;

    case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8':
        light_count = key - '0';
        refreshLights();
        return GL_TRUE;

    case 'S':
        {
        KDchar message[128];
        KDchar logValue[10];
        int polygons = Leaves_polyCount() +
                       Branches_polyCount() +
                       Ground_polyCount();

        kdStrcpy_s(message, 128, "frames per second ");
        kdLtostr(logValue, 10, (int)fps);
        kdStrncat_s(message, 128, logValue, 10);
        NvGl2DemoLogMessage(message);

        kdStrcpy_s(message, 128, "leaves ");
        kdLtostr(logValue, 10, Leaves_leafCount());
        kdStrncat_s(message, 128, logValue, 10);
        NvGl2DemoLogMessage(message);

        kdStrcpy_s(message, 128, "branches ");
        kdLtostr(logValue, 10, Branches_branchCount());
        kdStrncat_s(message, 128, logValue, 10);
        NvGl2DemoLogMessage(message);

        kdStrcpy_s(message, 128, "polygons per frame ");
        kdLtostr(logValue, 10, polygons);
        kdStrncat_s(message, 128, logValue, 10);
        NvGl2DemoLogMessage(message);

        kdStrcpy_s(message, 128, "polygons per second ");
        kdLtostr(logValue, 10, (int)(polygons * fps));
        kdStrncat_s(message, 128, logValue, 10);
        NvGl2DemoLogMessage(message);

        return GL_TRUE;
        }

    case 'V':
    case 'v':
        {
            KDchar buf[10];
            swapInterval += (key == 'V' ? 1 : -1);
            eglSwapInterval(gfxState.display, swapInterval);
            NvGl2DemoLogMessage("swap interval set to:");
            kdLtostr(buf,10, swapInterval);
            NvGl2DemoLogMessage(buf);
            return GL_TRUE;
        }
    case 'r':
        Tree_toggleVBO();
        return GL_TRUE;
    }
    return GL_FALSE;
}


void Screen_callback(int key, int x, int y)
{
    if (overlay_flag && sliderControl(key)) {
        return;
    }
    if (!overlay_flag && cameraControl(key)) {
        return;
    }
    sceneControl(key);
}

void Screen_draw()
{
    const float s = 0.05f;
    double aspect = ((double)width)/((double)height);
    float  to_h;
    int    i, j;
    float  h, p, sh, ch, sp, cp;
    float3 forward_vec, tmp;
    float  scenemvp[16];
    float  treemvp[16];

    // Update the clock
    tick();

    // Clear the buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //
    // Adjust the camera so view direction shifts towards origin
    //

    velocity = velocity * 0.95f;

    dp = dp * 0.8f;

    to_h = (float)(kdAtan2f(-eye[0], eye[2]) * 180.0f/KD_PI_F);
    dh = (to_h - heading);
    while (dh > 180.0f) {
        dh -= 360.0f;
    }
    while (dh <= -180.0f) {
        dh += 360.0f;
    }
    dh *= 20.0f;

    heading += (float)(dh*dt*0.05f);
    pitch += (float)(dp*dt*0.05f);

    h = degToRadF(heading);
    p = degToRadF(pitch);
    sh = kdSinf(h);
    ch = kdCosf(h);
    sp = kdSinf(p);
    cp = kdCosf(p);

    forward_vec[0] = sh*cp;
    forward_vec[1] = sp;
    forward_vec[2] = -ch*cp;

    mult_f3f(tmp, forward_vec, velocity);
    multi_f3f(tmp, (float)dt);
    addi_f3(eye, tmp);

    /* clamps */
    if (eye[1] < 0.5f) {
        eye[1] = 0.5f;
    }
    if (pitch > 80.0f) {
        pitch = 80.0f;
    }
    if (pitch < -80.0f) {
        pitch = -80.0f;
    }

    // Set up modelview/projection matrix for current camera settings
    NvGl2DemoMatrixIdentity(scenemvp);
    NvGl2DemoMatrixFrustum(scenemvp, -s * ((float) aspect), s * ((float) aspect),
                             -s, s,
                             0.1f, 1000.0f);
    NvGl2DemoMatrixRotate(scenemvp, pitch, -1.0f, 0.0f, 0.0f);
    NvGl2DemoMatrixRotate(scenemvp, heading, 0.0f, 1.0f, 0.0f);
    NvGl2DemoMatrixTranslate(scenemvp, -eye[0], -eye[1], -eye[2]);
    NvGl2DemoMatrixRotate(scenemvp, -90, 1.0f, 0.0f, 0.0f);

    // Enable depth testing for the scene
    glEnable(GL_DEPTH_TEST);

    // Render the trees and the ground beneath them
    for(j = 0; j < tree_poses.elem_count; j++) {
        TreePos *tree_pos = (TreePos *)Array_get(&tree_poses, j);
        Firefly* fireflies = tree_pos->fireflies;

        // Update firefly positions
        for (i=0; i<light_count; ++i) {
            Firefly_move(fireflies + i);
        }

        // Adjust modelview/projection for tree position/orientation
        kdMemcpy(treemvp, scenemvp, sizeof(treemvp));
        NvGl2DemoMatrixTranslate(treemvp, tree_pos->x, tree_pos->y, 0.0f);
        NvGl2DemoMatrixRotate(treemvp, tree_pos->angle, 0.0f, 0.0f, 1.0f);

        // Set up branch/ground shader
        glUseProgram(prog_solids);
        glUniformMatrix4fv(uloc_solidsMvpMat, 1, GL_FALSE, treemvp);
        glUniform3fv(uloc_solidsLightPos, light_count, fPos);
        glUniform3fv(uloc_solidsLightCol, light_count, fColor);

        // Set up leaf shader
        glUseProgram(prog_leaves);
        glUniformMatrix4fv(uloc_leavesMvpMat, 1, GL_FALSE, treemvp);
        glUniform3fv(uloc_leavesLightPos, light_count, fPos);
        glUniform3fv(uloc_leavesLightCol, light_count, fColor);

        // Render the tree
        Tree_draw();
    }

    // Draw the sky
    if (!nosky) {
        glUseProgram(prog_simplecol);
        glUniformMatrix4fv(uloc_simplecolMvpMat, 1, GL_FALSE, scenemvp);
        glUseProgram(prog_simpletex);
        glUniformMatrix4fv(uloc_simpletexMvpMat, 1, GL_FALSE, scenemvp);
        Sky_draw();
    }

    // Fireflies must be rendered after the rest of the scene because
    //   they disable depth mask so as not to self-occlude.
    glUseProgram(prog_simplecol);
    for (j = 0; j < tree_poses.elem_count; j++) {
        TreePos *tree_pos = (TreePos *)Array_get(&tree_poses, j);
        Firefly* fireflies = tree_pos->fireflies;

        // Adjust modelview/projection for tree position/orientation
        kdMemcpy(treemvp, scenemvp, sizeof(treemvp));
        NvGl2DemoMatrixTranslate(treemvp, tree_pos->x, tree_pos->y, 0.0f);
        NvGl2DemoMatrixRotate(treemvp, tree_pos->angle, 0.0f, 0.0f, 1.0f);
        glUniformMatrix4fv(uloc_simplecolMvpMat, 1, GL_FALSE, treemvp);

        // Draw fireflies
        for (i=0; i<light_count; ++i) {
            Firefly_draw(fireflies + i);
        }
    }

    // Disable depth testing and turn on blendin for the overlay
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

    // Do the sliders and logo
    if (overlay_flag) {
        int i;
        for(i = 0; i < NUM_TREE_PARAMS; i++) {
            Slider_draw(sliders[i]);
        }
        for(i = 0; i < NUM_PICTS; i++) {
            Picture_draw(picts[i]);
        }
    } else {
        Picture_draw(picts[LOGO_PICT]);
    }

    if (nvtxf && fps_flag) {
        KDchar buf[30];
        kdSnprintfKHR(buf, 30, "fps : %.1f", fps);
        nvtexfont2RenderString_All(nvtxf, buf, 0.4f, -0.9f, 0.5f, 1.0f, 1.0f, 1.0f);
    }

    glDisable(GL_BLEND);
}

void Screen_setDemoParams() {
    setSlider(TREE_PARAM_DEPTH,       0.56f,   GL_TRUE);
    setSlider(TREE_PARAM_BRANCH_SIZE, 0.1588f, GL_TRUE);
    setSlider(TREE_PARAM_FULLNESS,    0.62f,   GL_TRUE);
    light_count = 8;
    fps_flag = GL_FALSE;
    overlay_flag = GL_FALSE;
    swapInterval = 1;
    eglSwapInterval(gfxState.display, swapInterval);
}

void Screen_setSmallTex(void) {
    smalltex = 1;
}

void Screen_setNoSky(void) {
    nosky = 1;
}

void Screen_setNoMenu(void) {
    nomenu = 1;
    overlay_flag = GL_FALSE;
    fps_flag = GL_FALSE;
}

GLboolean Screen_isFinished() {
    return (duration > 0.0) && (currentTime - startTime >= duration);
}
