#include <KD/kd.h>
#include "shaders.h"
#include "nvgl2demo_common.h"

// When using binary shaders, we compile them in so the application
//   doesn't need to load extra data files
#ifdef USE_BINARY_SHADERS
static const char shad_lightingVert[] = {
#   include "lighting_vert.cghex"
};
static const char shad_solidsFrag[] = {
#   include "solids_frag.cghex"
};
static const char shad_leavesFrag[] = {
#   include "leaves_frag.cghex"
};
static const char shad_simplecolVert[] = {
#   include "simplecol_vert.cghex"
};
static const char shad_simplecolFrag[] = {
#   include "simplecol_frag.cghex"
};
static const char shad_simpletexVert[] = {
#   include "simpletex_vert.cghex"
};
static const char shad_simpletexFrag[] = {
#   include "simpletex_frag.cghex"
};
static const char shad_overlaycolVert[] = {
#   include "overlaycol_vert.cghex"
};
static const char shad_overlaycolFrag[] = {
#   include "overlaycol_frag.cghex"
};
static const char shad_overlaytexVert[] = {
#   include "overlaytex_vert.cghex"
};
static const char shad_overlaytexFrag[] = {
#   include "overlaytex_frag.cghex"
};
#endif

// Ground and branch shader (lit objects with full opacity)
GLint prog_solids = 0;
GLint uloc_solidsLights;
GLint uloc_solidsLightPos;
GLint uloc_solidsLightCol;
GLint uloc_solidsMvpMat;
GLint uloc_solidsTexUnit;
GLint aloc_solidsVertex;
GLint aloc_solidsNormal;
GLint aloc_solidsColor;
GLint aloc_solidsTexcoord;

// Leaves shader (lit objects with alphatest)
GLint prog_leaves = 0;
GLint uloc_leavesLights;
GLint uloc_leavesLightPos;
GLint uloc_leavesLightCol;
GLint uloc_leavesMvpMat;
GLint uloc_leavesTexUnit;
GLint aloc_leavesVertex;
GLint aloc_leavesNormal;
GLint aloc_leavesColor;
GLint aloc_leavesTexcoord;

// Simple colored object shader
GLint prog_simplecol = 0;
GLint uloc_simplecolMvpMat;
GLint aloc_simplecolVertex;
GLint aloc_simplecolColor;

// Simple textured object shader
GLint prog_simpletex = 0;
GLint uloc_simpletexColor;
GLint uloc_simpletexMvpMat;
GLint uloc_simpletexTexUnit;
GLint aloc_simpletexVertex;
GLint aloc_simpletexTexcoord;

// Colored overlay shader
GLint prog_overlaycol = 0;
GLint aloc_overlaycolVertex;
GLint aloc_overlaycolColor;

// Textured overlay shader
GLint prog_overlaytex = 0;
GLint uloc_overlaytexTexUnit;
GLint aloc_overlaytexVertex;
GLint aloc_overlaytexTexcoord;

// Load all the shaders and extract uniform/attribute locations
int LoadShaders(void)
{
    GLboolean success;

#   ifdef USE_BINARY_SHADERS
    // If using binary shaders, create the programs from the hardcoded headers
    prog_solids =
        NvGl2DemoLoadShaderBinaries(
            shad_lightingVert, sizeof(shad_lightingVert),
            shad_solidsFrag,   sizeof(shad_solidsFrag),
            GL_TRUE);
    prog_leaves =
        NvGl2DemoLoadShaderBinaries(
            shad_lightingVert, sizeof(shad_lightingVert),
            shad_leavesFrag,   sizeof(shad_leavesFrag),
            GL_TRUE);
    prog_simplecol =
        NvGl2DemoLoadShaderBinaries(
            shad_simplecolVert, sizeof(shad_simplecolVert),
            shad_simplecolFrag, sizeof(shad_simplecolFrag),
            GL_TRUE);
    prog_simpletex =
        NvGl2DemoLoadShaderBinaries(
            shad_simpletexVert, sizeof(shad_simpletexVert),
            shad_simpletexFrag, sizeof(shad_simpletexFrag),
            GL_TRUE);
    prog_overlaycol =
        NvGl2DemoLoadShaderBinaries(
            shad_overlaycolVert, sizeof(shad_overlaycolVert),
            shad_overlaycolFrag, sizeof(shad_overlaycolFrag),
            GL_TRUE);
    prog_overlaytex =
        NvGl2DemoLoadShaderBinaries(
            shad_overlaytexVert, sizeof(shad_overlaytexVert),
            shad_overlaytexFrag, sizeof(shad_overlaytexFrag),
            GL_TRUE);
#   else
    // Otherwise, load and compile the shader sources
    prog_solids =
        NvGl2DemoLoadShaderSources(
            "lighting_vert.glslv", "solids_frag.glslf", GL_FALSE);
    prog_leaves =
        NvGl2DemoLoadShaderSources(
            "lighting_vert.glslv", "leaves_frag.glslf", GL_FALSE);
    prog_simplecol =
        NvGl2DemoLoadShaderSources(
            "simplecol_vert.glslv", "simplecol_frag.glslf", GL_FALSE);
    prog_simpletex =
        NvGl2DemoLoadShaderSources(
            "simpletex_vert.glslv", "simpletex_frag.glslf", GL_FALSE);
    prog_overlaycol =
        NvGl2DemoLoadShaderSources(
            "overlaycol_vert.glslv", "overlaycol_frag.glslf", GL_FALSE);
    prog_overlaytex =
        NvGl2DemoLoadShaderSources(
            "overlaytex_vert.glslv", "overlaytex_frag.glslf", GL_FALSE);
#   endif
    success =  prog_solids && prog_leaves
            && prog_simplecol  && prog_simpletex
            && prog_overlaycol && prog_overlaytex;
    if (!success) {
        NvGl2DemoLogMessage("Error occured loading shaders");
        return 0;
    }

    // Load locations for branch/ground shader
    uloc_solidsLights   = glGetUniformLocation(prog_solids, "lights");
    uloc_solidsLightPos = glGetUniformLocation(prog_solids, "lightpos");
    uloc_solidsLightCol = glGetUniformLocation(prog_solids, "lightcol");
    uloc_solidsMvpMat   = glGetUniformLocation(prog_solids, "mvpmatrix");
    uloc_solidsTexUnit  = glGetUniformLocation(prog_solids, "texunit");
    aloc_solidsVertex   = glGetAttribLocation(prog_solids,  "vertex");
    aloc_solidsNormal   = glGetAttribLocation(prog_solids,  "normal");
    aloc_solidsColor    = glGetAttribLocation(prog_solids,  "color");
    aloc_solidsTexcoord = glGetAttribLocation(prog_solids,  "texcoord");
    success =  (uloc_solidsLights   >= 0)
            && (uloc_solidsLightPos >= 0)
            && (uloc_solidsLightCol >= 0)
            && (uloc_solidsMvpMat   >= 0)
            && (uloc_solidsTexUnit  >= 0)
            && (aloc_solidsVertex   >= 0)
            && (aloc_solidsNormal   >= 0)
            && (aloc_solidsColor    >= 0)
            && (aloc_solidsTexcoord >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving branch/ground shader locations");
        return 0;
    }

    // Load locations for leaves shader
    uloc_leavesLights   = glGetUniformLocation(prog_leaves, "lights");
    uloc_leavesLightPos = glGetUniformLocation(prog_leaves, "lightpos");
    uloc_leavesLightCol = glGetUniformLocation(prog_leaves, "lightcol");
    uloc_leavesMvpMat   = glGetUniformLocation(prog_leaves, "mvpmatrix");
    uloc_leavesTexUnit  = glGetUniformLocation(prog_leaves, "texunit");
    aloc_leavesVertex   = glGetAttribLocation(prog_leaves,  "vertex");
    aloc_leavesNormal   = glGetAttribLocation(prog_leaves,  "normal");
    aloc_leavesColor    = glGetAttribLocation(prog_leaves,  "color");
    aloc_leavesTexcoord = glGetAttribLocation(prog_leaves,  "texcoord");
    success =  (uloc_leavesLights   >= 0)
            && (uloc_leavesLightPos >= 0)
            && (uloc_leavesLightCol >= 0)
            && (uloc_leavesMvpMat   >= 0)
            && (uloc_leavesTexUnit  >= 0)
            && (aloc_leavesVertex   >= 0)
            && (aloc_leavesNormal   >= 0)
            && (aloc_leavesColor    >= 0)
            && (aloc_leavesTexcoord >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving leaves shader locations");
        return 0;
    }

    // Load locations for simple color shader
    uloc_simplecolMvpMat = glGetUniformLocation(prog_simplecol, "mvpmatrix");
    aloc_simplecolVertex = glGetAttribLocation(prog_simplecol, "vertex");
    aloc_simplecolColor  = glGetAttribLocation(prog_simplecol, "color");
    success =  (aloc_simplecolColor  >= 0)
            && (uloc_simplecolMvpMat >= 0)
            && (aloc_simplecolVertex >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving simple color shader locations");
        return 0;
    }

    // Load locations for simple texture shader
    uloc_simpletexColor    = glGetUniformLocation(prog_simpletex, "color");
    uloc_simpletexMvpMat   = glGetUniformLocation(prog_simpletex, "mvpmatrix");
    uloc_simpletexTexUnit  = glGetUniformLocation(prog_simpletex, "texunit");
    aloc_simpletexVertex   = glGetAttribLocation(prog_simpletex,  "vertex");
    aloc_simpletexTexcoord = glGetAttribLocation(prog_simpletex,  "texcoord");
    success =  (uloc_simpletexTexUnit  >= 0)
            && (uloc_simpletexMvpMat   >= 0)
            && (uloc_simpletexColor    >= 0)
            && (aloc_simpletexVertex   >= 0)
            && (aloc_simpletexTexcoord >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving simple texture shader locations");
        return 0;
    }

    // Load locations for colored overlay shader
    aloc_overlaycolVertex = glGetAttribLocation(prog_overlaycol, "vertex");
    aloc_overlaycolColor  = glGetAttribLocation(prog_overlaycol, "color");
    success =  (aloc_overlaycolColor  >= 0)
            && (aloc_overlaycolVertex >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving colored overlay shader locations");
        return 0;
    }

    // Load locations for textured overlay shader
    uloc_overlaytexTexUnit  = glGetUniformLocation(prog_overlaytex, "texunit");
    aloc_overlaytexVertex   = glGetAttribLocation(prog_overlaytex, "vertex");
    aloc_overlaytexTexcoord = glGetAttribLocation(prog_overlaytex, "texcoord");
    success =  (uloc_overlaytexTexUnit  >= 0)
            && (aloc_overlaytexVertex   >= 0)
            && (aloc_overlaytexTexcoord >= 0);
    if (!success) {
        NvGl2DemoLogMessage("Error occured retrieving textured overlay shader locations");
        return 0;
    }

    return 1;
}

void FreeShaders(void)
{
    if (prog_overlaytex) glDeleteProgram(prog_overlaytex);
    if (prog_overlaycol) glDeleteProgram(prog_overlaycol);
    if (prog_simpletex)  glDeleteProgram(prog_simpletex);
    if (prog_simplecol)  glDeleteProgram(prog_simplecol);
    if (prog_leaves)     glDeleteProgram(prog_leaves);
    if (prog_solids)     glDeleteProgram(prog_solids);
}
