#include <KD/kd.h>
#include "vector.h"

#include "slider.h"
#include "shaders.h"


/* local functions.  */
void Slider_build_rod(Slider *o);
void Slider_build_knob(Slider *o);
void Slider_build_colors(Slider *o);


static int rod_geom[6][2] = {
    {4, 7},
    {4, 9},
    {2, 9},
    {4, 11},
    {2, 11},
    {4, 13}
};

static int knob_polys[5][4] = {
    {0, 1, 3, 2},
    {2, 3, 5, 4},
    {4, 5, 7, 6},
    {6, 7, 1, 0},
    {1, 3, 5, 7}
};

static int knob_geom[8][2] = {
    {0, 2},
    {0, 6},
    {-4, 10},
    {-2, 10},
    {0, 18},
    {0, 14},
    {4, 10},
    {2, 10}
};

Slider *Slider_new(float mn, float mx, float init)
{
    Slider *o = (Slider *)kdMalloc(sizeof(Slider));
    o->minimum = mn;
    o->maximum = mx;
    o->value = init;
    o->mouse_down = GL_FALSE;

    set_3(o->rod_color, 0.2f, 0.1f, 0.8f);
    set_3(o->knob_color, 1.0f, 0.2f, 0.4f);
    Slider_build_colors(o);

    return o;
}

void Slider_delete(Slider *o)
{
    kdFree(o);
}

void Slider_build_rod(Slider *o)
{
    float scale = (o->top-o->bottom)/20.0f;
    int i;
    for (i=0; i<6; ++i) {
        float x = rod_geom[i][0]*scale;
        float y = rod_geom[i][1]*scale + o->bottom;

        set_2(o->rod_vertices[i], x+o->left, y);
        set_2(o->rod_vertices[i+6], o->right-x, y);

    }
}

void Slider_build_colors(Slider *o)
{
    {
        float3 white = {1.0, 1.0, 1.0};
        float3 dark, medium;
        div_f3(dark, o->rod_color, 3.0f);
        copy_3(medium, o->rod_color);

        copy_3(o->rod_colors[0], dark);
        copy_3(o->rod_colors[1], medium);
        copy_3(o->rod_colors[2], dark);
        copy_3(o->rod_colors[3], white);
        copy_3(o->rod_colors[4], dark);
        copy_3(o->rod_colors[5], medium);

        copy_3(o->rod_colors[6], dark);
        copy_3(o->rod_colors[7], medium);
        copy_3(o->rod_colors[8], medium);
        copy_3(o->rod_colors[9], white);
        copy_3(o->rod_colors[10], medium);
        copy_3(o->rod_colors[11], medium);
    }

    {
        float3 white = {1.0, 1.0, 1.0};
        float3 medium_dark, dark, medium, light;
        div_f3(medium_dark, o->knob_color, 2.0f);
        div_f3(dark, o->knob_color, 3.0f);
        copy_3(medium, o->knob_color);
        add_f3(light, white, o->knob_color);
        divi_f3(light, 2.0f);


        copy_3(o->knob_colors[0], dark);
        copy_3(o->knob_colors[1], light);
        copy_3(o->knob_colors[2], white);
        copy_3(o->knob_colors[3], medium_dark);
        copy_3(o->knob_colors[4], medium);
    }
}


void Slider_build_knob(Slider *o)
{
    float scale = (o->top-o->bottom)/20.0f;
    float length = (o->right-o->left)/scale;

    float n = (o->value - o->minimum)/(o->maximum-o->minimum);
    int  i;

    o->knob_pos = 4 + (length-8) * n;
    for (i=0; i<8; ++i) {
        float x = (knob_geom[i][0]+o->knob_pos)*scale+o->left;
        float y = knob_geom[i][1]*scale+o->bottom;
        set_2(o->knob_vertices[i], x, y);
    }
}


void Slider_setPos(Slider *o, float l, float r, float b, float t)
{
    o->left = l;
    o->right = r;
    o->bottom = b;
    o->top = t;

    Slider_build_knob(o);
    Slider_build_rod(o);
}


void Slider_draw(Slider *o)
{
    unsigned short indices[8] = {0,6,1,7,3,9,5,11};
    int            k;

    // Load shader program and enable vertex array
    glUseProgram(prog_overlaycol);
    glEnableVertexAttribArray(aloc_overlaycolVertex);

    // Draw slider
    glEnableVertexAttribArray(aloc_overlaycolColor);
    glVertexAttribPointer(aloc_overlaycolVertex,
                          2, GL_FLOAT, GL_FALSE, 0, o->rod_vertices);
    glVertexAttribPointer(aloc_overlaycolColor,
                          3, GL_FLOAT, GL_FALSE, 0, o->rod_colors);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 6);
    glDrawArrays(GL_TRIANGLE_STRIP, 6, 6);
    glDrawElements(GL_TRIANGLE_STRIP, 8, GL_UNSIGNED_SHORT, indices);
    glDisableVertexAttribArray(aloc_overlaycolColor);

    // Draw knob
    glVertexAttribPointer(aloc_overlaycolVertex,
                          2, GL_FLOAT, GL_FALSE, 0, o->knob_vertices);
    for (k=0; k<5; ++k) {
        glVertexAttrib3fv(aloc_overlaycolColor, o->knob_colors[k]);
        glDrawElements(GL_TRIANGLE_FAN, 4, GL_UNSIGNED_INT, knob_polys[k]);
    }

    // Disable vertex array
    glDisableVertexAttribArray(aloc_overlaycolVertex);
}


GLboolean Slider_setValue(Slider *o, float val)
{
    float oldval = o->value;
    o->value = val;
    if (o->value>o->maximum) o->value = o->maximum;
    if (o->value<o->minimum) o->value = o->minimum;
    if (oldval == o->value)
        return GL_FALSE;
    Slider_build_knob(o);
    return GL_TRUE;
}

float Slider_getValue(Slider *o) { return o->value; }

void Slider_select(Slider *o, GLboolean sel)
{
    if (sel) {
        set_3(o->rod_color, 0.0f, 0.9f, 0.2f);
        set_3(o->knob_color, 1.0f, 0.2f, 0.4f);
    } else {
        set_3(o->rod_color, 0.2f, 0.1f, 0.8f);
        set_3(o->knob_color, 1.0f, 0.2f, 0.4f);
    }
    Slider_build_colors(o);
}
