/*
 * nvgl2demo_common_x11.c
 *
 * Copyright 2009, NVIDIA Corporation.
 * All rights Reserved.
 *
 * THE INFORMATION CONTAINED HEREIN IS PROPRIETARY AND CONFIDENTIAL TO
 * NVIDIA, CORPORATION.  USE, REPRODUCTION OR DISCLOSURE TO ANY THIRD PARTY
 * IS SUBJECT TO WRITTEN PRE-APPROVAL BY NVIDIA, CORPORATION.
 */

//
// This file contains utility functions that we end up replicating in most
//   of our demos. It would be cleaner to have a single copy of this
//   somewhere, but replicating it makes the demos more self-contained for
//   teaching purposes.
//
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/stat.h>
#include <nvgl2demo_common.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>

struct NvGl2DemoPlatformState
{
    Display* XDisplay;
    int XScreen;
};

static NvGl2DemoPlatformState presetPlatformState;

#define MAX_ATTRIB 1  //maximum number of attributes passed to eglCreateWindowSurface
///////////////////////////////////////////////////////////////////////////////
// Prints a message to standard out
void NvGl2DemoLogMessage(const char* message, ...)
{
    va_list ap;
    int length;

    length = strlen(message);
    if (length > 0) {
        va_start(ap, message);
        vprintf(message,ap);
        va_end(ap);

        // if not newline terminated, add a newline
        if (message[length-1] != '\n') {
            printf("\n");
         }
    }
}
//
// Argument parsing
//

// Extract an argument from the list
static const char* NvGl2DemoArgExtract(int* argc, char** argv, int index)
{
    char* val = argv[index];
    int   j;
    for (j=index+1; j<*argc; ++j) {
        argv[j-1] = argv[j];
    }
    (*argc)--;
    return val;
}

#if defined(EGL_TRIPLE_BUFFER_NV) || defined(EGL_QUADRUPLE_BUFFER_NV)
// locate a substring in a null terminated string, return NULL if not found
static char* NvGl2DemoStrStr(const char* str1, const char* str2)
{
    int len1;
    int len2;
    char* start;

    if (!str1 || !str2)
    {
        return NULL;
    }

    len1 = strlen(str1);
    len2 = strlen(str2);

    // Find the first occurence
    start = strchr(str1, *str2);

    // If we found the first character, continue to test
    // See if is possible to fit str2 into the rest of str1
    while ( start ) {

        if (!strncmp(str2, start, len2)) {
            return start;
        }

        // Find the first occurence
        start = strchr(start + 1, *str2);
    }

    // If we got here, we didn't find the string
    return NULL;
}
#endif

///////////////////////////////////////////////////////////////////////////////

//
// Starting up and shutting down window system and EGL
//

// egl/window system state
NvGl2DemoGfxState gfxState =
{
    (NativeDisplayType)0,
    (NativeWindowType)0,
    EGL_NO_DISPLAY,
    EGL_NO_SURFACE,
    (EGLConfig)0,
    EGL_NO_CONTEXT,
    0,
    0,
    2,
    0,
    NULL,
};

// return gl2demo usage string under x
char * NvGl2DemoGetUsageString(void)
{
  return "  [-res <xres> <yres>] \n"
         "  [-csaa <samples (5 supported on Tegra)>] \n"
         "  [-vcaa <samples (5 supported on Tegra)>] (identical to -csaa)\n"
         "  [-fsaa <samples (4 supported on Tegra)>]\n";
}

// Start up, initializing native window system and EGL
int NvGl2DemoInitialize(int* argc, char** argv, const char *appName,
                        EGLint eglApiBit)
{
    // NOTE: This array is modified by the code below. The indices used
    // are hardcoded numbers, so be careful if you change the order of
    // the values!
    EGLint configAttrs[] = {
        EGL_RED_SIZE,        1,
        EGL_GREEN_SIZE,      1,
        EGL_BLUE_SIZE,       1,
        EGL_DEPTH_SIZE,     16,
        EGL_RENDERABLE_TYPE, EGL_NONE,  // NOTE: value gets overwritten
        EGL_SAMPLES,         0,
#ifdef EGL_NV_coverage_sample
        EGL_COVERAGE_BUFFERS_NV,  0,
        EGL_COVERAGE_SAMPLES_NV,  0,
#endif
        EGL_NONE
    };
    EGLint contextAttrs[] = {
        EGL_NONE, 0,                    // NOTE: value gets overwritten
                                        // for GL
        EGL_NONE
    };
    EGLenum eglApi;

    int        windowSize[2]  = { 0, 0 };
    EGLConfig* configList = NULL;
    EGLint     configCount;
    EGLBoolean eglStatus;
    EGLint     buffering = 0;
    EGLint     csaa = 0;
    EGLint     fsaa = 0;
    EGLint     windowAttrs[2*MAX_ATTRIB + 1];
    EGLint     windowIndex = 0;  //index of first kdFree place in windowAttrs
    int        i;
    const char* query;
    XSizeHints size_hints;
    Atom       wm_destroy_window;

    // Validate input API parameter and translate it to API enumerant
    switch (eglApiBit) {
        case EGL_OPENGL_ES2_BIT:
            eglApi = EGL_OPENGL_ES_API;
            break;

        case EGL_OPENVG_BIT:
            eglApi = EGL_OPENVG_API;
            break;

        default:
            printf("Invalid/unsupported API requested\n");
            goto fail;
    }

    // Parse input parameters
    for (i=1; i<*argc; /*nop*/) {
        if (!strncmp(argv[i], "-csaa", 5)) {
            argv[i] += 5;
            if (argv[i][0] == 0) NvGl2DemoArgExtract(argc, argv, i);
            if (i == *argc) {
                printf("Missing parameter for -csaa\n");
                goto fail;
            }
            csaa = strtol(argv[i], NULL, 10);
            if (csaa != 5) {
                printf("Invalid parameter(%s) for -csaa\n", argv[i]);
                goto fail;
            }
            NvGl2DemoArgExtract(argc, argv, i);
        } else if (!strncmp(argv[i], "-vcaa", 5)) {
            argv[i] += 5;
            if (argv[i][0] == 0) NvGl2DemoArgExtract(argc, argv, i);
            if (i == *argc) {
                printf("Missing parameter for -vcaa\n");
                goto fail;
            }
            csaa = strtol(argv[i], NULL, 10);
            if (csaa != 5) {
                printf("Invalid parameter(%s) for -vcaa\n", argv[i]);
                goto fail;
            }
            NvGl2DemoArgExtract(argc, argv, i);
        } else if (!strncmp(argv[i], "-fsaa", 5)) {
            argv[i] += 5;
            if (argv[i][0] == 0) NvGl2DemoArgExtract(argc, argv, i);
            if (i == *argc) {
                printf("Missing parameter for -fsaa\n");
                goto fail;
            }
            fsaa = strtol(argv[i], NULL, 10);
            if (fsaa != 4) {
                printf("Invalid parameter(%s) for -fsaa\n", argv[i]);
                goto fail;
            }
            NvGl2DemoArgExtract(argc, argv, i);
        }
        else if (!strncmp(argv[i],"-buffering",10)) {
            argv[i] += 10;
            NvGl2DemoArgExtract(argc, argv, i);
            if (argv[i][0] == 0) NvGl2DemoArgExtract(argc, argv, i);
            if (i == *argc) {
                break;
            }

            buffering = strtol(argv[i],NULL,10);
            if(buffering < 0 || buffering > 4) {
                printf("Invalid parameter(%s) for -buffering", argv[i]);
                goto fail;
            }
            NvGl2DemoArgExtract(argc, argv, i);
        }
        else if (!strncmp(argv[i],"-res",4)) {
            argv[i] += 4;
            if (argv[i][0] == 0) NvGl2DemoArgExtract(argc, argv, i);
            if (i == *argc) {
                printf("Missing parameters for -res."
                            " Need to follow with xres yres.\n");
                goto fail;
            }

            windowSize[0] = strtol(argv[i],NULL,10);
            NvGl2DemoArgExtract(argc, argv, i);
            windowSize[1] = strtol(argv[i],NULL,10);
            NvGl2DemoArgExtract(argc, argv, i);
        }
        else {
            i++;
        }
    }

    gfxState.platformspecific = (NvGl2DemoPlatformState*)malloc(sizeof(NvGl2DemoPlatformState));
    if (!gfxState.platformspecific) {
        printf("Could not allocate platform specific storage memory\n");
        goto fail;
    }


    // Open X display
    if (presetPlatformState.XDisplay) {
        gfxState.platformspecific->XDisplay = presetPlatformState.XDisplay;
    } else {
        gfxState.platformspecific->XDisplay = XOpenDisplay(NULL);
        if (gfxState.platformspecific->XDisplay == NULL) {
            printf("X failed to open display\n");
            goto fail;
        }
    }

    // Obtain the EGL display
    gfxState.display = eglGetDisplay((EGLNativeDisplayType) gfxState.platformspecific->XDisplay);
    if (gfxState.display == EGL_NO_DISPLAY) {
        printf("EGL failed to obtain display\n");
        goto fail;
    }

    // Initialize EGL
    eglStatus = eglInitialize(gfxState.display, 0, 0);
    if (!eglStatus) {
        printf("EGL failed to initialize\n");
        goto fail;
    }

    // If FSAA requested and supported, get a config that supports it.
    if (fsaa) configAttrs[11] = fsaa;

    // If CSAA requested and supported, get a config that supports it
    if (csaa) {
        query = eglQueryString(gfxState.display, EGL_EXTENSIONS);
        if(!strstr(query, " EGL_NV_coverage_sample")) {
            printf("CSAA (VCAA) not supported!\n");
            goto fail;
        }
        configAttrs[13]  = 1;
        configAttrs[15]  = csaa;
    }

    // Put API-specific values into the config and context attributes.
    // This assumes that the EGL_RENDERABLE_TYPE enum is the fourth one
    // in the configAttrs array and that the contextAttrs array is
    // empty but has room for an EGL_CONTEXT_CLIENT_VERSION enum.
    //
    // Also, while we're at it, bind the client API.
    configAttrs[9] = eglApiBit;
    switch (eglApiBit) {
        case EGL_OPENGL_ES2_BIT:
            contextAttrs[0] = EGL_CONTEXT_CLIENT_VERSION;
            contextAttrs[1] = 2;
            break;

        default:
            break;
    }
    eglBindAPI(eglApi);

    // Find out how many configurations suit our needs
    eglStatus = eglChooseConfig(gfxState.display, configAttrs,
                                NULL, 0, &configCount);
    if (!eglStatus || !configCount) {
        printf("EGL failed to return any matching configurations\n");
        goto fail;
    }

    // Allocate room for the list of matching configurations
    configList = (EGLConfig*)malloc(configCount * sizeof(EGLConfig));
    if (!configList) {
        printf("malloc failure obtaining configuration list\n");
        goto fail;
    }

    // Obtain the configuration list from EGL
    eglStatus = eglChooseConfig(gfxState.display, configAttrs,
                                configList, configCount, &configCount);
    if (!eglStatus || !configCount) {
        printf("EGL failed to populate configuration list\n");
        goto fail;
    }

    // Select an EGL configuration that matches the native window
    gfxState.config = configList[0];

    // Set buffering options
    if(buffering) {

        gfxState.buffering = buffering;

        query = eglQueryString(gfxState.display, EGL_EXTENSIONS);
        windowAttrs[windowIndex++] = EGL_RENDER_BUFFER;

        switch(buffering) {
            case 1:
                windowAttrs[windowIndex++] = EGL_SINGLE_BUFFER;
                break;

            case 2:
                windowAttrs[windowIndex++] = EGL_BACK_BUFFER;
                break;

            #ifdef EGL_TRIPLE_BUFFER_NV
            case 3:
                if(!NvGl2DemoStrStr(query, "EGL_NV_triple_buffer ")) {
                    printf("TRIPLE_BUFFER not supported!\n");
                    goto fail;
                }
                windowAttrs[windowIndex++] = EGL_TRIPLE_BUFFER_NV;
                break;
            #endif

            #ifdef EGL_QUADRUPLE_BUFFER_NV
            case 4:
                if(!NvGl2DemoStrStr(query, "EGL_NV_quadruple_buffer ")) {
                    printf("QUADRUPLE_BUFFER not supported!\n");
                    goto fail;
                }
                windowAttrs[windowIndex++] = EGL_QUADRUPLE_BUFFER_NV;
                break;
            #endif

            default:
                printf("Choosen buffering not defined.\n");
                goto fail;
        }
    }

    // Set window size. If no specified, default to 800x480
    if (!windowSize[0]) windowSize[0] = 800;
    if (!windowSize[1]) windowSize[1] = 480;

    if (presetPlatformState.XScreen) {
        gfxState.platformspecific->XScreen = presetPlatformState.XScreen;
    } else {
        gfxState.platformspecific->XScreen =
            DefaultScreen(gfxState.platformspecific->XDisplay);
    }

    // Create a native window
    gfxState.nativeWindow = (NativeWindowType)
        XCreateSimpleWindow(gfxState.platformspecific->XDisplay,
                            RootWindow(gfxState.platformspecific->XDisplay,
                                       gfxState.platformspecific->XScreen),
                            0, 0, windowSize[0], windowSize[1], 0,
                            BlackPixel(gfxState.platformspecific->XDisplay,
                                       gfxState.platformspecific->XScreen),
                            WhitePixel(gfxState.platformspecific->XDisplay,
                                       gfxState.platformspecific->XScreen));
    if (!gfxState.nativeWindow) {
        printf("Error creating native window\n");
        goto fail;
    }

    size_hints.flags = PPosition | PSize | PMinSize;
    size_hints.x = 0;
    size_hints.y = 0;
    size_hints.min_width  = (windowSize[0] < 320) ? windowSize[0] : 320;
    size_hints.min_height = (windowSize[1] < 240) ? windowSize[1] : 240;
    size_hints.width  = windowSize[0];
    size_hints.height = windowSize[1];

    XSetStandardProperties(gfxState.platformspecific->XDisplay,
                           (Window) gfxState.nativeWindow,
                           appName, appName, None,
                           argv, *argc, &size_hints);

    XSetWindowBackgroundPixmap(gfxState.platformspecific->XDisplay,
                               (Window) gfxState.nativeWindow,
                               None);

    wm_destroy_window = XInternAtom(gfxState.platformspecific->XDisplay,
                                    "WM_DELETE_WINDOW", False);
    XSetWMProtocols(gfxState.platformspecific->XDisplay,
                    (Window) gfxState.nativeWindow,
                    &wm_destroy_window,
                    True);

    // Make sure the MapNotify event goes into the queue
    XSelectInput(gfxState.platformspecific->XDisplay,
                 (Window) gfxState.nativeWindow,
                 StructureNotifyMask);

    XMapWindow(gfxState.platformspecific->XDisplay,
               (Window) gfxState.nativeWindow);

    // Create an EGL window surface for the native window
    windowAttrs[windowIndex++] = EGL_NONE;
    gfxState.surface = eglCreateWindowSurface(gfxState.display,
                                              gfxState.config,
                                              gfxState.nativeWindow,
                                              windowAttrs);
    if (!gfxState.surface) {
        printf("EGL couldn't create window\n");
        goto fail;
    }

    // Create an EGL context
    gfxState.context = eglCreateContext(gfxState.display,
                                        gfxState.config,
                                        NULL,
                                        contextAttrs);
    if (!gfxState.context) {
        printf("EGL couldn't create context\n");
        goto fail;
    }

    // Make the context and surface current for rendering
    eglStatus = eglMakeCurrent(gfxState.display,
                               gfxState.surface, gfxState.surface,
                               gfxState.context);
    if (!eglStatus) {
        printf("EGL couldn't make context/surface current\n");
        goto fail;
    }

    // Store the rendering API away
    gfxState.api = eglApi;

    // Query the EGL surface width and height
    eglStatus =  eglQuerySurface(gfxState.display, gfxState.surface,
                                 EGL_WIDTH,  &gfxState.width)
              && eglQuerySurface(gfxState.display, gfxState.surface,
                                 EGL_HEIGHT, &gfxState.height);
    if (!eglStatus) {
        printf("EGL couldn't get window width/height\n");
        goto fail;
    }

    // free the configuration list and return success
    free(configList);
    return 1;

    // On failure, clean up partial initialization
    fail:
    if (configList) free(configList);
    NvGl2DemoShutdown();
    return 0;
}


void *NvGl2DemoCreatePixmap(unsigned int width, unsigned int height,
                            unsigned int depth)
{
    Pixmap ret;
    ret = XCreatePixmap(gfxState.platformspecific->XDisplay,
                        (Window) gfxState.nativeWindow,
                        width, height, depth);
    if (!ret) {
        printf("Error allocating X11 pixmap\n");
    }
    return (void *) ret;
}


int NvGl2DemoPresetPlatformState(const char *state, void *value)
{
    if (!strcmp("Display", state)) {
        presetPlatformState.XDisplay = (Display *) value;
    } else if (!strcmp("Screen", state)) {
        presetPlatformState.XScreen = (int) value;
    } else {
        printf("Unknown X11 platform state (%s)\n", state);
        return 0;
    }
    return 1;
}


// Shut down, kdFreeing all EGL and native window system resources.
void NvGl2DemoShutdown(void)
{
    EGLBoolean eglStatus;

    // Clear rendering context
    // Note that we need to bind the API to unbind... yick
    if (gfxState.display != EGL_NO_DISPLAY) {
        eglBindAPI(gfxState.api);
        eglStatus = eglMakeCurrent(gfxState.display,
                                   EGL_NO_SURFACE, EGL_NO_SURFACE,
                                   EGL_NO_CONTEXT);
        if (!eglStatus)
            printf("Error clearing current surfaces/context\n");
    }

    // Destroy the EGL context
    if (gfxState.context != EGL_NO_CONTEXT) {
        eglStatus = eglDestroyContext(gfxState.display, gfxState.context);
        if (!eglStatus)
            printf("Error destroying EGL context\n");
        gfxState.context = EGL_NO_CONTEXT;
    }

    // Destroy the EGL surface
    if (gfxState.surface != EGL_NO_SURFACE) {
        eglStatus = eglDestroySurface(gfxState.display, gfxState.surface);
        if (!eglStatus)
            printf("Error destroying EGL surface\n");
        gfxState.surface = EGL_NO_SURFACE;
    }

    // Terminate EGL
    if (gfxState.display != EGL_NO_DISPLAY) {
        eglStatus = eglTerminate(gfxState.display);
        if (!eglStatus)
            printf("Error terminating EGL\n");
        gfxState.display = EGL_NO_DISPLAY;
    }

    // Close the native window
    if (gfxState.platformspecific && gfxState.nativeWindow) {
        XDestroyWindow(gfxState.platformspecific->XDisplay,
                       (Window) gfxState.nativeWindow);
        gfxState.nativeWindow = (NativeWindowType)0;
        XCloseDisplay(gfxState.platformspecific->XDisplay);
        free(gfxState.platformspecific);
        gfxState.platformspecific = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

//
// Callback handling
//
static NvGl2DemoCloseCB   closeCB   = NULL;
static NvGl2DemoResizeCB  resizeCB  = NULL;
static NvGl2DemoKeyCB     keyCB     = NULL;
static NvGl2DemoPointerCB pointerCB = NULL;
static NvGl2DemoButtonCB  buttonCB     = NULL;
static long               eventMask = 0;

static void NvGl2DemoUpdateEventMask(void)
{
    eventMask = StructureNotifyMask;
    if (keyCB) eventMask |= (KeyPressMask | KeyReleaseMask);
    if (pointerCB) eventMask |= (PointerMotionMask | PointerMotionHintMask);
    if (buttonCB) eventMask |= (ButtonPressMask | ButtonReleaseMask);
    XSelectInput(gfxState.platformspecific->XDisplay,
                 (Window) gfxState.nativeWindow,
                 eventMask);
}

void NvGl2DemoSetCloseCB(NvGl2DemoCloseCB cb)
{
    closeCB  = cb;
    NvGl2DemoUpdateEventMask();
}

void NvGl2DemoSetResizeCB(NvGl2DemoResizeCB cb)
{
    resizeCB = cb;
    NvGl2DemoUpdateEventMask();
}

void NvGl2DemoSetKeyCB(NvGl2DemoKeyCB cb)
{
    keyCB = cb;
    NvGl2DemoUpdateEventMask();
}

void NvGl2DemoSetPointerCB(NvGl2DemoPointerCB cb)
{
    pointerCB = cb;
    NvGl2DemoUpdateEventMask();
}

void NvGl2DemoSetButtonCB(NvGl2DemoButtonCB cb)
{
    buttonCB = cb;
    NvGl2DemoUpdateEventMask();
}

void NvGl2DemoCheckEvents(void)
{
    XEvent event;
    KeySym key;
    char   str[2] = {0,0};
    Window root, child;
    int    rootX, rootY;
    int    winX, winY;
    unsigned int mask;

    while (XPending(gfxState.platformspecific->XDisplay)) {
        XNextEvent(gfxState.platformspecific->XDisplay, &event);
        switch (event.type) {

            case ConfigureNotify:
                if (!resizeCB) break;
                if (  (event.xconfigure.width  != gfxState.width)
                   || (event.xconfigure.height != gfxState.height) ) {
                    resizeCB(event.xconfigure.width, event.xconfigure.height);
                    gfxState.width = event.xconfigure.width;
                    gfxState.height = event.xconfigure.height;
                }
                break;

            case ClientMessage:
                if (event.xclient.message_type ==
                        XInternAtom(gfxState.platformspecific->XDisplay,
                                    "WM_PROTOCOLS", True) &&
                    event.xclient.data.l[0] ==
                        XInternAtom(gfxState.platformspecific->XDisplay,
                                    "WM_DELETE_WINDOW", True) &&
                    closeCB)
                    closeCB();
                break;

            case KeyPress:
            case KeyRelease:
                if (!keyCB) break;
                XLookupString(&event.xkey, str, 2, &key, NULL);
                if (!str[0] || str[1]) break; // Only care about basic keys
                keyCB(str[0], (event.type == KeyPress) ? 1 : 0);
                break;

            case MotionNotify:
                if (!pointerCB) break;
                XQueryPointer(gfxState.platformspecific->XDisplay,
                              event.xmotion.window,
                              &root, &child,
                              &rootX, &rootY, &winX, &winY, &mask);
                pointerCB(winX, winY);
                break;

            case ButtonPress:
            case ButtonRelease:
                if (!buttonCB) break;
                buttonCB(event.xbutton.button,
                         (event.type == ButtonPress) ? 1 : 0);
                break;

            case MapNotify:
                // Once the window is mapped, can set the focus
                XSetInputFocus(gfxState.platformspecific->XDisplay,
                               event.xmap.window,
                               RevertToNone,
                               CurrentTime);
                break;

           default:
               break;
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

//
// Shader program creation
//

// Loads a file containing shader code
static char* NvGl2DemoLoadFile(const char *file)
{
    const char *filename;
    char path[1024];
    int filesize;
    char *data;
    struct stat st;
    FILE *f;

    // Look in both the current and parent dir
    sprintf(path, "../%s", file);
    if(!(f = fopen(filename = path + 1, "rb")) &&
       !(f = fopen(filename = path, "rb"))) {
        printf("Unable to open shader: %s\n", file);
        return 0;
    }

    stat(filename, &st);
    filesize = st.st_size;
    if (filesize == -1) {
        printf("Unable to get shader file size: %s\n", filename);
        fclose(f);
        return 0;
    }

    data = (char *)malloc(filesize + 1);
    if(!data) {
        printf("Unable to alloc memory to read shader: %s\n", filename);
        fclose(f);
        return 0;
    }
    if(fread(data, 1, filesize, f) != filesize) {
        printf("Unable to read shader: %s\n", filename);
        free(data);
        fclose(f);
        return 0;
    }
    data[filesize] = 0;
    fclose(f);

    return data;
}

// If shader compilation fails, prints the associated logs
static void NvGl2DemoShaderDebug(GLuint obj, GLenum status, const char* op)
{
    int success;
    // log output.
    int len;
    char *str = NULL;
    if (status == GL_COMPILE_STATUS) {
        glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &len);
        if (len > 0) {
            str = (char *)malloc(len * sizeof(char));
            glGetShaderInfoLog(obj, len, NULL, str);
        }
    } else { // LINK or VALIDATE
        glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &len);
        if (len > 0) {
            str = (char *)malloc(len * sizeof(char));
            glGetProgramInfoLog(obj, len, NULL, str);
        }
    }
    if (str != NULL && *str != '\0') { printf("--- %s log ---\n%s", op, str); }
    if (str) { free(str); }

    // check the compile / link status.
    if (status == GL_COMPILE_STATUS) {
        glGetShaderiv(obj, status, &success);
        if (!success) {
            glGetShaderiv(obj, GL_SHADER_SOURCE_LENGTH, &len);
            if (len > 0) {
                str = (char *)malloc(len * sizeof(char));
                glGetShaderSource(obj, len, NULL, str);
                if (str != NULL && *str != '\0') {
                    printf("--- %s code ---\n%s", op, str); }
                free(str);
            }
        }
    } else { // LINK or VALIDATE
        glGetProgramiv(obj, status, &success);
    }

    if (!success)
    {
        printf("--- %s failed\n", op);
        exit(-1);
    }
}

// Take precompiled shader binaries and builds a shader program
GLuint NvGl2DemoLoadShaderBinaries(
    const char* vertBin, GLuint vertBinSize,
    const char* fragBin, GLuint fragBinSize,
    GLboolean debugging)
{
// Binary shaders not supportable for non-ES OpenGL
#ifdef GL_ES_VERSION_2_0

    GLuint prog;
    GLuint vertShader;
    GLuint fragShader;

    // Create the program
    prog = glCreateProgram();

    // Create the GL shader objects
    vertShader = glCreateShader(GL_VERTEX_SHADER);
    fragShader = glCreateShader(GL_FRAGMENT_SHADER);

    // Load the binary data into the shader objects
    glShaderBinary(1, &vertShader,
                   GL_NVIDIA_PLATFORM_BINARY_NV, vertBin, vertBinSize);
    glShaderBinary(1, &fragShader,
                   GL_NVIDIA_PLATFORM_BINARY_NV, fragBin, fragBinSize);

    // Attach the shaders to the program
    glAttachShader(prog, vertShader);
    glAttachShader(prog, fragShader);

    // Delete the shaders
    glDeleteShader(vertShader);
    glDeleteShader(fragShader);

    // Link and validate the shader program
    glLinkProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_LINK_STATUS, "Program Link");

    return prog;
#else  // GL_ES_VERSION_2_0
    return 0;
#endif // GL_ES_VERSION_2_0
}

// Takes shader source files, compiles them, and builds a shader program
GLuint NvGl2DemoLoadShaderSources(
    const char* vertFile,
    const char* fragFile,
    GLboolean debugging)
{
    GLuint prog = 0;
    char*  vertSource;
    char*  fragSource;
    GLint  vertSourceLen;
    GLint  fragSourceLen;
    GLuint vertShader;
    GLuint fragShader;

    // Load the shader files
    vertSource    = NvGl2DemoLoadFile(vertFile);
    fragSource    = NvGl2DemoLoadFile(fragFile);
    if (!vertSource || !fragSource) goto done;
    vertSourceLen = (GLint)strlen(vertSource);
    fragSourceLen = (GLint)strlen(fragSource);

    // Create the program
    prog = glCreateProgram();

    // Create the GL shader objects
    vertShader = glCreateShader(GL_VERTEX_SHADER);
    fragShader = glCreateShader(GL_FRAGMENT_SHADER);

    // Load shader sources into GL and compile
    glShaderSource(vertShader, 1, (const char**)&vertSource, &vertSourceLen);
    glCompileShader(vertShader);
    if (debugging)
        NvGl2DemoShaderDebug(vertShader, GL_COMPILE_STATUS, "Vert Compile");
    glShaderSource(fragShader, 1, (const char**)&fragSource, &fragSourceLen);
    glCompileShader(fragShader);
    if (debugging)
        NvGl2DemoShaderDebug(fragShader, GL_COMPILE_STATUS, "Frag Compile");

    // Attach the shaders to the program
    glAttachShader(prog, vertShader);
    glAttachShader(prog, fragShader);

    // Delete the shaders
    glDeleteShader(vertShader);
    glDeleteShader(fragShader);

    // Link and validate the shader program
    glLinkProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_LINK_STATUS, "Program Link");
    glValidateProgram(prog);
    if (debugging)
        NvGl2DemoShaderDebug(prog, GL_VALIDATE_STATUS, "Program Validate");

    done:

    free(fragSource);
    free(vertSource);
    return prog;
}

///////////////////////////////////////////////////////////////////////////////

// Math stuff

int eq(float a, float b)  {
    return fabs(a-b) <= eps;
}

//
// Matrix functions, since GLES 2.x doesn't provide them
//

void NvGl2DemoMatrixIdentity(float m[16])
{
    memset(m, 0, sizeof(float) * 16);
    m[4 * 0 + 0] = m[4 * 1 + 1] = m[4 * 2 + 2] = m[4 * 3 + 3] = 1.0;
}

int NvGl2DemoMatrixEquals(float a[16], float b[16])  {
    int i;
    for(i = 0; i < 16; ++i)
        if(!eq(a[i], b[i]))
            return 0;

    return 1;
}

void NvGl2DemoMatrixTranspose(float m[16])
{
    int i, j;
    float t;
    for(i = 1; i < 4; ++i)
        for(j = 0; j < i; ++j)  {
            t = m[4*i+j];
            m[4*i+j] = m[4*j+i];
            m[4*j+i] = t;
        }
}

void NvGl2DemoMatrixMultiply(float m0[16], float m1[16])
{
    int r, c, i;
    for(r = 0; r < 4; r++)
    {
        float m[4] = {0.0, 0.0, 0.0, 0.0};
        for(c = 0; c < 4; c++)
        {
            for(i = 0; i < 4; i++)
            {
                m[c] += m0[4 * i + r] * m1[4 * c + i];
            }
        }
        for(c = 0; c < 4; c++) { m0[4 * c + r] = m[c]; }
    }
}

void NvGl2DemoMatrixMultiply_4x4_3x3(float m0[16], float m1[9])
{
    int r, c, i;
    for(r = 0; r < 4; r++)
    {
        float m[3] = {0.0, 0.0, 0.0};
        for(c = 0; c < 3; c++)
        {
            for(i = 0; i < 3; i++)
            {
                m[c] += m0[4 * i + r] * m1[3 * c + i];
            }
        }
        for(c = 0; c < 3; c++) { m0[4 * c + r] = m[c]; }
    }
}

void NvGl2DemoMatrixMultiply_3x3(float m0[9], float m1[9])
{
    int r, c, i;
    for(r = 0; r < 3; r++)
    {
        float m[3] = {0.0, 0.0, 0.0};
        for(c = 0; c < 3; c++)
        {
            for(i = 0; i < 3; i++)
            {
                m[c] += m0[3 * i + r] * m1[3 * c + i];
            }
        }
        for(c = 0; c < 3; c++) { m0[3 * c + r] = m[c]; }
    }
}

void NvGl2DemoMatrixFrustum(float m[16],
                    float l, float r, float b, float t, float n, float f)
{
    float m1[16];
    float rightMinusLeftInv, topMinusBottomInv, farMinusNearInv, twoNear;

    rightMinusLeftInv = 1.0f / (r - l);
    topMinusBottomInv = 1.0f / (t - b);
    farMinusNearInv = 1.0f / (f - n);
    twoNear = 2.0f * n;

    m1[ 0] = twoNear * rightMinusLeftInv;
    m1[ 1] = 0.0f;
    m1[ 2] = 0.0f;
    m1[ 3] = 0.0f;

    m1[ 4] = 0.0f;
    m1[ 5] = twoNear * topMinusBottomInv;
    m1[ 6] = 0.0f;
    m1[ 7] = 0.0f;

    m1[ 8] = (r + l) * rightMinusLeftInv;
    m1[ 9] = (t + b) * topMinusBottomInv;
    m1[10] = -(f + n) * farMinusNearInv;
    m1[11] = -1.0f;

    m1[12] = 0.0f;
    m1[13] = 0.0f;
    m1[14] = -(twoNear * f) * farMinusNearInv;
    m1[15] = 0.0f;

    NvGl2DemoMatrixMultiply(m, m1);
}

void NvGl2DemoMatrixOrtho(float m[16],
                  float l, float r, float b, float t, float n, float f)
{
    float m1[16];
    float rightMinusLeftInv, topMinusBottomInv, farMinusNearInv;

    rightMinusLeftInv = 1.0f / (r - l);
    topMinusBottomInv = 1.0f / (t - b);
    farMinusNearInv = 1.0f / (f - n);

    m1[ 0] = 2.0f * rightMinusLeftInv;
    m1[ 1] = 0.0f;
    m1[ 2] = 0.0f;
    m1[ 3] = 0.0f;

    m1[ 4] = 0.0f;
    m1[ 5] = 2.0f * topMinusBottomInv;
    m1[ 6] = 0.0f;
    m1[ 7] = 0.0f;

    m1[ 8] = 0.0f;
    m1[ 9] = 0.0f;
    m1[10] = -2.0f * farMinusNearInv;
    m1[11] = 0.0f;

    m1[12] = -(r + l) * rightMinusLeftInv;
    m1[13] = -(t + b) * topMinusBottomInv;
    m1[14] = -(f + n) * farMinusNearInv;
    m1[15] = 1.0f;

    NvGl2DemoMatrixMultiply(m, m1);
}

void NvGl2DemoMatrixTranslate(float m[16], float x, float y, float z)
{
    float m1[16];
    NvGl2DemoMatrixIdentity(m1);

    m1[4 * 3 + 0] = x;
    m1[4 * 3 + 1] = y;
    m1[4 * 3 + 2] = z;

    NvGl2DemoMatrixMultiply(m, m1);
}

void NvGl2DemoMatrixRotate_create3x3(float m[9],
                             float theta, float x, float y, float z)
{
    float len = (float)sqrt(x * x + y * y + z * z);
    float u0 = x / len;
    float u1 = y / len;
    float u2 = z / len;
    float rad = (float)(theta / 180 * M_PI);
    float c = (float)cos(rad);
    float s = (float)sin(rad);
    m[3 * 0 + 0] = u0 * u0 + c * (1 - u0 * u0) + s * 0;
    m[3 * 0 + 1] = u0 * u1 + c * (0 - u0 * u1) + s * u2;
    m[3 * 0 + 2] = u0 * u2 + c * (0 - u0 * u2) - s * u1;

    m[3 * 1 + 0] = u1 * u0 + c * (0 - u1 * u0) - s * u2;
    m[3 * 1 + 1] = u1 * u1 + c * (1 - u1 * u1) + s * 0;
    m[3 * 1 + 2] = u1 * u2 + c * (0 - u1 * u2) + s * u0;

    m[3 * 2 + 0] = u2 * u0 + c * (0 - u2 * u0) + s * u1;
    m[3 * 2 + 1] = u2 * u1 + c * (0 - u2 * u1) - s * u0;
    m[3 * 2 + 2] = u2 * u2 + c * (1 - u2 * u2) + s * 0;
}

void NvGl2DemoMatrixRotate(float m[16], float theta, float x, float y, float z)
{
    float r[9];
    NvGl2DemoMatrixRotate_create3x3(r, theta, x, y, z);
    NvGl2DemoMatrixMultiply_4x4_3x3(m, r);
}

void NvGl2DemoMatrixRotate_3x3(float m[9], float theta, float x, float y, float z)
{
    float r[9];
    NvGl2DemoMatrixRotate_create3x3(r, theta, x, y, z);
    NvGl2DemoMatrixMultiply_3x3(m, r);
}

float NvGl2DemoMatrixDeterminant(float m[16]) {
   return
       m[4*0+3] * m[4*1+2] * m[4*2+1] * m[4*3+0]-m[4*0+2] * m[4*1+3] * m[4*2+1] * m[4*3+0]-m[4*0+3] * m[4*1+1] * m[4*2+2] * m[4*3+0]+m[4*0+1] * m[4*1+3]    * m[4*2+2] * m[4*3+0]+
       m[4*0+2] * m[4*1+1] * m[4*2+3] * m[4*3+0]-m[4*0+1] * m[4*1+2] * m[4*2+3] * m[4*3+0]-m[4*0+3] * m[4*1+2] * m[4*2+0] * m[4*3+1]+m[4*0+2] * m[4*1+3]    * m[4*2+0] * m[4*3+1]+
       m[4*0+3] * m[4*1+0] * m[4*2+2] * m[4*3+1]-m[4*0+0] * m[4*1+3] * m[4*2+2] * m[4*3+1]-m[4*0+2] * m[4*1+0] * m[4*2+3] * m[4*3+1]+m[4*0+0] * m[4*1+2]    * m[4*2+3] * m[4*3+1]+
       m[4*0+3] * m[4*1+1] * m[4*2+0] * m[4*3+2]-m[4*0+1] * m[4*1+3] * m[4*2+0] * m[4*3+2]-m[4*0+3] * m[4*1+0] * m[4*2+1] * m[4*3+2]+m[4*0+0] * m[4*1+3]    * m[4*2+1] * m[4*3+2]+
       m[4*0+1] * m[4*1+0] * m[4*2+3] * m[4*3+2]-m[4*0+0] * m[4*1+1] * m[4*2+3] * m[4*3+2]-m[4*0+2] * m[4*1+1] * m[4*2+0] * m[4*3+3]+m[4*0+1] * m[4*1+2]    * m[4*2+0] * m[4*3+3]+
       m[4*0+2] * m[4*1+0] * m[4*2+1] * m[4*3+3]-m[4*0+0] * m[4*1+2] * m[4*2+1] * m[4*3+3]-m[4*0+1] * m[4*1+0] * m[4*2+2] * m[4*3+3]+m[4*0+0] * m[4*1+1]    * m[4*2+2] * m[4*3+3];
}

void NvGl2DemoMatrixInverse(float m[16]) {
    float a[16];
    float det;
    int i;
    float b[16], e[16];

    a[4*0+0] = m[4*1+2]*m[4*2+3]*m[4*3+1] - m[4*1+3]*m[4*2+2]*m[4*3+1] + m[4*1+3]*m[4*2+1]*m[4*3+2] - m[4*1+1]*m[4*2+3]*m[4*3+2] - m[4*1+2]*m[4*2+1]*m[4*3+3] + m[4*1+1]*m[4*2+2]*m[4*3+3];
    a[4*0+1] = m[4*0+3]*m[4*2+2]*m[4*3+1] - m[4*0+2]*m[4*2+3]*m[4*3+1] - m[4*0+3]*m[4*2+1]*m[4*3+2] + m[4*0+1]*m[4*2+3]*m[4*3+2] + m[4*0+2]*m[4*2+1]*m[4*3+3] - m[4*0+1]*m[4*2+2]*m[4*3+3];
    a[4*0+2] = m[4*0+2]*m[4*1+3]*m[4*3+1] - m[4*0+3]*m[4*1+2]*m[4*3+1] + m[4*0+3]*m[4*1+1]*m[4*3+2] - m[4*0+1]*m[4*1+3]*m[4*3+2] - m[4*0+2]*m[4*1+1]*m[4*3+3] + m[4*0+1]*m[4*1+2]*m[4*3+3];
    a[4*0+3] = m[4*0+3]*m[4*1+2]*m[4*2+1] - m[4*0+2]*m[4*1+3]*m[4*2+1] - m[4*0+3]*m[4*1+1]*m[4*2+2] + m[4*0+1]*m[4*1+3]*m[4*2+2] + m[4*0+2]*m[4*1+1]*m[4*2+3] - m[4*0+1]*m[4*1+2]*m[4*2+3];
    a[4*1+0] = m[4*1+3]*m[4*2+2]*m[4*3+0] - m[4*1+2]*m[4*2+3]*m[4*3+0] - m[4*1+3]*m[4*2+0]*m[4*3+2] + m[4*1+0]*m[4*2+3]*m[4*3+2] + m[4*1+2]*m[4*2+0]*m[4*3+3] - m[4*1+0]*m[4*2+2]*m[4*3+3];
    a[4*1+1] = m[4*0+2]*m[4*2+3]*m[4*3+0] - m[4*0+3]*m[4*2+2]*m[4*3+0] + m[4*0+3]*m[4*2+0]*m[4*3+2] - m[4*0+0]*m[4*2+3]*m[4*3+2] - m[4*0+2]*m[4*2+0]*m[4*3+3] + m[4*0+0]*m[4*2+2]*m[4*3+3];
    a[4*1+2] = m[4*0+3]*m[4*1+2]*m[4*3+0] - m[4*0+2]*m[4*1+3]*m[4*3+0] - m[4*0+3]*m[4*1+0]*m[4*3+2] + m[4*0+0]*m[4*1+3]*m[4*3+2] + m[4*0+2]*m[4*1+0]*m[4*3+3] - m[4*0+0]*m[4*1+2]*m[4*3+3];
    a[4*1+3] = m[4*0+2]*m[4*1+3]*m[4*2+0] - m[4*0+3]*m[4*1+2]*m[4*2+0] + m[4*0+3]*m[4*1+0]*m[4*2+2] - m[4*0+0]*m[4*1+3]*m[4*2+2] - m[4*0+2]*m[4*1+0]*m[4*2+3] + m[4*0+0]*m[4*1+2]*m[4*2+3];
    a[4*2+0] = m[4*1+1]*m[4*2+3]*m[4*3+0] - m[4*1+3]*m[4*2+1]*m[4*3+0] + m[4*1+3]*m[4*2+0]*m[4*3+1] - m[4*1+0]*m[4*2+3]*m[4*3+1] - m[4*1+1]*m[4*2+0]*m[4*3+3] + m[4*1+0]*m[4*2+1]*m[4*3+3];
    a[4*2+1] = m[4*0+3]*m[4*2+1]*m[4*3+0] - m[4*0+1]*m[4*2+3]*m[4*3+0] - m[4*0+3]*m[4*2+0]*m[4*3+1] + m[4*0+0]*m[4*2+3]*m[4*3+1] + m[4*0+1]*m[4*2+0]*m[4*3+3] - m[4*0+0]*m[4*2+1]*m[4*3+3];
    a[4*2+2] = m[4*0+1]*m[4*1+3]*m[4*3+0] - m[4*0+3]*m[4*1+1]*m[4*3+0] + m[4*0+3]*m[4*1+0]*m[4*3+1] - m[4*0+0]*m[4*1+3]*m[4*3+1] - m[4*0+1]*m[4*1+0]*m[4*3+3] + m[4*0+0]*m[4*1+1]*m[4*3+3];
    a[4*2+3] = m[4*0+3]*m[4*1+1]*m[4*2+0] - m[4*0+1]*m[4*1+3]*m[4*2+0] - m[4*0+3]*m[4*1+0]*m[4*2+1] + m[4*0+0]*m[4*1+3]*m[4*2+1] + m[4*0+1]*m[4*1+0]*m[4*2+3] - m[4*0+0]*m[4*1+1]*m[4*2+3];
    a[4*3+0] = m[4*1+2]*m[4*2+1]*m[4*3+0] - m[4*1+1]*m[4*2+2]*m[4*3+0] - m[4*1+2]*m[4*2+0]*m[4*3+1] + m[4*1+0]*m[4*2+2]*m[4*3+1] + m[4*1+1]*m[4*2+0]*m[4*3+2] - m[4*1+0]*m[4*2+1]*m[4*3+2];
    a[4*3+1] = m[4*0+1]*m[4*2+2]*m[4*3+0] - m[4*0+2]*m[4*2+1]*m[4*3+0] + m[4*0+2]*m[4*2+0]*m[4*3+1] - m[4*0+0]*m[4*2+2]*m[4*3+1] - m[4*0+1]*m[4*2+0]*m[4*3+2] + m[4*0+0]*m[4*2+1]*m[4*3+2];
    a[4*3+2] = m[4*0+2]*m[4*1+1]*m[4*3+0] - m[4*0+1]*m[4*1+2]*m[4*3+0] - m[4*0+2]*m[4*1+0]*m[4*3+1] + m[4*0+0]*m[4*1+2]*m[4*3+1] + m[4*0+1]*m[4*1+0]*m[4*3+2] - m[4*0+0]*m[4*1+1]*m[4*3+2];
    a[4*3+3] = m[4*0+1]*m[4*1+2]*m[4*2+0] - m[4*0+2]*m[4*1+1]*m[4*2+0] + m[4*0+2]*m[4*1+0]*m[4*2+1] - m[4*0+0]*m[4*1+2]*m[4*2+1] - m[4*0+1]*m[4*1+0]*m[4*2+2] + m[4*0+0]*m[4*1+1]*m[4*2+2];

    det = NvGl2DemoMatrixDeterminant(m);

    for(i = 0; i < 16; ++i)
        a[i] /= det;

    NvGl2DemoMatrixIdentity(e);

    NvGl2DemoMatrixCopy(b, m);
    NvGl2DemoMatrixMultiply(b, a);

    NvGl2DemoMatrixCopy(m, a);
}

void NvGl2DemoMatrixCopy(float dest[16], float src[16])  {
    memcpy(dest, src, 16*sizeof(float));
}

void NvGl2DemoMatrixPrint(float a[16])  {
    int i, j;

    for(i = 0; i < 4; ++i)
        for(j = 0; j < 4; ++j)
            printf("%f%c", a[4*i+j], j == 3 ? '\n' : ' ');
}

void NvGl2DemoMatrixVectorMultiply(float m[16],  float v[4])  {
    float res[4];
    int i, j;

    for(i = 0; i < 4; ++i)  {
        res[i] = 0;
        for(j = 0; j < 4; ++j)
            res[i] += m[i*4 + j] * v[j];
    }

    memcpy(v, res, sizeof(res));
}
