#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/stat.h>
#include <nvgl2demo_common.h>
#include <stdlib.h>
#include <stdio.h>

// gluBuild2DMipmaps() from mglu.c was used in original C++ code
// It's been slightly cut for OGLES2
static int gluBuild2DMipmaps(GLenum target, GLint internalFormat,
                             GLsizei width, GLsizei height,
                             GLenum format, GLenum type, const void *data)
{
    int lod, x, y, c;
    int pitch=0, shift, components;
    int comp[4];
    GLubyte *buf, *dst;
    const GLubyte *src1, *src2;

    // Handle some error cases
    if ((width < 1 || height < 1) ||                 // 0 width/height
        ((width&(width-1) || height&(height-1))) ||  // power of two check
        (type != GL_UNSIGNED_BYTE)) {                // must be UNSIGNED_BYTE
        return -1;
    }

    switch (format) {
        case GL_RGBA: components = 4; break;
        case GL_RGB: components = 3; break;
        case GL_LUMINANCE_ALPHA: components = 2; break;
        case GL_LUMINANCE:
        case GL_ALPHA: components = 1; break;
        default: components = 0; break;
    }

    // Reserve space for lod 1, (width/2) * (height/2) * 4, which is just...
    // For RGB or BGR, we reserve extra space.  So what?
    buf = malloc(width*height);

    lod = 0;
    do {
        // Download a level
        glTexImage2D(target, lod, internalFormat, width, height, 0,
                     format, GL_UNSIGNED_BYTE, data);
        // After the 1x1, we're done.
        if (width == 1 && height == 1) {
            break;
        }

        // Observation:
        // Normal MxN downsampling will average a 2x2 block of texels into 1.
        // Nx1 or 1xN downsampling will average 2 successive texels into 1.
        // The Nx1 and 1xN cases can be handled without distinction between
        // width and height.
        dst = buf;
        src1 = src2 = data;
        // Nx1 or 1xN case
        if (width == 1 || height == 1) {
            shift = 1;
        }
        // General MxN case
        else {
            shift = 2;
            pitch = width*components;
            src2 += pitch;
        }

        if (width > 1)  width >>= 1;
        if (height > 1) height >>= 1;

        for (y=0; y<height; y++) {
            for (x=0; x<width; x++) {
                for (c=0; c<components; c++) {
                    comp[c] = *src1++;
                }
                for (c=0; c<components; c++) {
                    comp[c] += *src1++;
                }
                if (shift == 2) {
                    for (c=0; c<components; c++) {
                        comp[c] += *src2++;
                    }
                    for (c=0; c<components; c++) {
                        comp[c] += *src2++;
                    }
                }
                // Write the downsampled texel.
                for (c=0; c<components; c++) {
                    *dst++ = comp[c]>>shift;
                }
            }
            // For MxN case, skip a row.
            if (shift == 2) {
                src1 += pitch;
                src2 += pitch;
            }
        }

        data = buf;
        lod++;
    } while (1);

    free(buf);
    return 0;
}


static FILE* OpenFile(const char *name)
{
    const char *filename;
    char path[1024];
    FILE *fp;

    // Look in both the current and parent dir
    sprintf(path, "../%s", name);
    if(!(fp = fopen(filename = path + 1, "rb")) &&
       !(fp = fopen(filename = path, "rb"))) {
        printf("Unable to open file: %s\n", name);
        return 0;
    }
    return fp;
}


static void tga_read_header(FILE *fp, const char *path, unsigned int *width, unsigned int *height, unsigned int *bpp, unsigned int *format)
{
    unsigned char header[18];
	
    fseek(fp, 0, SEEK_SET);
    fread(header, 1, 18, fp);
    if( (header[1] != 0) || (header[2] != 2) )
    {
    	printf("\nThe following file is not supported:%s", path);
       	printf("\nOnly uncompressed tga files are supported");
        exit(-1);    	
    }

    *width = header[13] << 8 | header[12];
    *height = header[15] << 8 | header[14];
	
    *bpp = header[16] >> 3;
    *format = (*bpp == 4)?GL_RGBA:GL_RGB;
}


static unsigned char* tga_read_body(FILE *fp, unsigned int width, unsigned int height, unsigned int bpp)
{
    int i, j;
    unsigned char* buffer;
    unsigned char* temp;
	
    buffer = (unsigned char *) malloc(width*height*bpp);
    if(!buffer)
    {
    	printf("\nMemory unavailable");
    	fclose(fp);
        exit(-1);    	
    }

    fseek(fp, 18, SEEK_SET);
    fread(buffer, 1, width*height*bpp, fp);

    temp = buffer;


    //BGRA -> RGBA
    for(i = 0; i < height; i++)
    {
        for(j = 0; j < width; j++)
        {
            unsigned char c;
    	
            c = temp[0];
            temp[0] = temp[2];
            temp[2] = c;
    		
            temp += bpp;
        }
    }

    return buffer;
}


//WARNING:
//This function cannot be used for all kinds of TGA files.
//Only uncompressed RGB/RGBA bitmaps are supported
GLuint NvGl2DemoLoadTga(const char* name)
{
    GLuint id;
    FILE *fp;
    unsigned char* buffer;
    unsigned int width, height;
    unsigned int bpp;//bytes per pixel
    unsigned int format;//pixel format... either GL_RGB or GL_RGBA


    fp = OpenFile(name);	
    tga_read_header(fp, name, &width, &height, &bpp, &format);
    buffer = tga_read_body(fp, width, height, bpp);
	
    glGenTextures(1, &id);
    glBindTexture(GL_TEXTURE_2D, id);
    gluBuild2DMipmaps(GL_TEXTURE_2D, format, width, height, format, GL_UNSIGNED_BYTE, buffer);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

    free(buffer);
    fclose(fp);

    return id;
}
